from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Menu
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import icon_menu_tags
from .permissions import permission_tag_create, permission_tag_view

menu_tags = Menu(
    condition=factory_condition_queryset_access(
        app_label='tags', model_name='Tag',
        object_permission=permission_tag_view,
        view_permission=permission_tag_create,
    ), icon=icon_menu_tags, label=_('Tags'), name='tags'
)



from mayan.apps.testing.tests.base import BaseTestCase

from ..exceptions import (
    DynamicSearchInterpreterUnknownSearchType, DynamicSearchScopedQueryError
)
from ..literals import (
    SCOPE_OPERATOR_AND, SCOPE_OPERATOR_OR, QUERY_PARAMETER_ANY_FIELD
)
from ..scoped_queries import (
    ScopedQueryEntryControlResult, ScopedQueryEntryDataFilter,
    ScopedQueryEntryDataOperator
)
from ..search_interpreters import (
    SearchInterpreter, SearchInterpreterAdvanced, SearchInterpreterScoped
)

from .mixins.search_interpreter_mixins import SearchInterpreterTestMixin
from .mixins.base import SearchTestMixin


class SearchInterpreterAdvancedDecodeTestCase(
    SearchInterpreterTestMixin, SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_advanced_special_any_fields_term_decode(self):
        kwargs = {
            'query': {QUERY_PARAMETER_ANY_FIELD: 'term1'},
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 1)
        self.assertEqual(len(scoped_query.scope_entry_list), 2)
        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            QUERY_PARAMETER_ANY_FIELD
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, 'term1'
        )
        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].result_scope_identifier,
            '0'
        )

    def test_advanced_special_any_fields_term_empty_decode(self):
        kwargs = {
            'query': {QUERY_PARAMETER_ANY_FIELD: ''},
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 0)
        self.assertEqual(len(scoped_query.scope_entry_list), 0)
        self.assertTrue(search_interpreter.is_empty)

    def test_advanced_match_all_off_decode(self):
        kwargs = {
            'query': {
                'field_1': 'term_1', 'field_2': 'term_2',
                '_match_all': 'false'
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 3)
        self.assertEqual(len(scoped_query.scope_entry_list), 4)

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, 'term_1'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].field_name,
            'field_2'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].scope_identifier, '1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].value, 'term_2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[2],
                ScopedQueryEntryDataOperator
            )
        )

        self.assertEqual(
            tuple(scoped_query.scope_entry_list[2].operand_list),
            ('0', '1')
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].operator_text,
            SCOPE_OPERATOR_OR
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].scope_identifier, '2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[3],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[3].result_scope_identifier,
            '2'
        )

    def test_advanced_match_all_on_decode(self):
        kwargs = {
            'query': {
                'field_1': 'term_1', 'field_2': 'term_2',
                '_match_all': 'true'
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 3)
        self.assertEqual(len(scoped_query.scope_entry_list), 4)

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, 'term_1'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].field_name,
            'field_2'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].scope_identifier, '1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].value, 'term_2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[2],
                ScopedQueryEntryDataOperator
            )
        )

        self.assertEqual(
            tuple(scoped_query.scope_entry_list[2].operand_list),
            ('0', '1')
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].operator_text,
            SCOPE_OPERATOR_AND
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].scope_identifier, '2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[3],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[3].result_scope_identifier,
            '2'
        )

    def test_advanced_literal_decode(self):
        kwargs = {
            'query': {
                'field_1': 'term_1 AND term_2',
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 3)
        self.assertEqual(len(scoped_query.scope_entry_list), 4)

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, 'term_1'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].scope_identifier, '1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].value, 'term_2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[2],
                ScopedQueryEntryDataOperator
            )
        )

        self.assertEqual(
            tuple(scoped_query.scope_entry_list[2].operand_list),
            ('0', '1')
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].operator_text,
            SCOPE_OPERATOR_AND
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].scope_identifier, '2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[3],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[3].result_scope_identifier,
            '2'
        )

    def test_advanced_literal_single_decode(self):
        kwargs = {
            'query': {
                'field_1': '"term_1 AND term_2"',
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 1)
        self.assertEqual(len(scoped_query.scope_entry_list), 2)

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, True
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, 'term_1 AND term_2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].result_scope_identifier,
            '0'
        )

    def test_advanced_literal_dual_with_operator_decode(self):
        kwargs = {
            'query': {
                'field_1': '"term_1" AND "term_2"',
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 3)
        self.assertEqual(len(scoped_query.scope_entry_list), 4)

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, True
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, 'term_1'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, True
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].scope_identifier, '1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].value, 'term_2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[2],
                ScopedQueryEntryDataOperator
            )
        )

        self.assertEqual(
            tuple(scoped_query.scope_entry_list[2].operand_list),
            ('0', '1')
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].operator_text,
            SCOPE_OPERATOR_AND
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].scope_identifier, '2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[3],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[3].result_scope_identifier,
            '2'
        )


class SearchInterpreterAdvancedDetectionTestCase(
    SearchInterpreterTestMixin, SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_search_interpreter_property_is_empty_from_empty_query(self):
        kwargs = {
            'query': {self._test_search_field.field_name: ''},
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        self.assertTrue(search_interpreter.is_empty)

    def test_subclass_advanced_selection(self):
        kwargs = {
            'query': {self._test_search_field.field_name: ''},
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        self.assertTrue(
            isinstance(search_interpreter, SearchInterpreterAdvanced)
        )
        self.assertTrue(search_interpreter.is_empty)

    def test_subclass_advanced_selection_from_any_field_query(self):
        kwargs = {
            'query': {QUERY_PARAMETER_ANY_FIELD: ''},
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        self.assertTrue(
            isinstance(search_interpreter, SearchInterpreterAdvanced)
        )
        self.assertTrue(search_interpreter.is_empty)


class SearchInterpreterCommonTestCase(
    SearchInterpreterTestMixin, SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_search_interpreter_property_is_empty(self):
        kwargs = {
            'query': {},
            'search_model': self._test_search_model
        }

        with self.assertRaises(expected_exception=DynamicSearchInterpreterUnknownSearchType):
            SearchInterpreter.init(**kwargs)


class SearchInterpreterScopedDecodeTestCase(
    SearchInterpreterTestMixin, SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_base_functionality(self):
        kwargs = {
            'query': {
                '__0_{}'.format('field_1'): 'term_1',
                '__1_{}'.format('field_2'): 'term_2',
                '__2': 'AND_0_1',
                '__result': '2'
            },
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 3)
        self.assertEqual(len(scoped_query.scope_entry_list), 4)

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, 'term_1'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].field_name,
            'field_2'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].scope_identifier, '1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].value, 'term_2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[2],
                ScopedQueryEntryDataOperator
            )
        )

        self.assertEqual(
            tuple(scoped_query.scope_entry_list[2].operand_list),
            ('0', '1')
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].operator_text,
            SCOPE_OPERATOR_AND
        )
        self.assertEqual(
            scoped_query.scope_entry_list[2].scope_identifier, '2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[3],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[3].result_scope_identifier,
            '2'
        )

    def test_multiple_terms(self):
        kwargs = {
            'query': {
                '__0_{}'.format('field_1'): 'term_1 term_2',
                '__result': '0'
            },
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            search_interpreter.do_query_decode()

    def test_quoted_term(self):
        kwargs = {
            'query': {
                '__0_{}'.format('field_1'): '"term_1 term_2"',
                '__result': '0'
            },
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 1)
        self.assertEqual(len(scoped_query.scope_entry_list), 2)

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, True
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, 'term_1 term_2'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].result_scope_identifier,
            '0'
        )

    def test_literal_dual_term(self):
        kwargs = {
            'query': {
                '__0_{}'.format('field_1'): '"term_1" "term_2"',
                '__result': '0'
            },
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            search_interpreter.do_query_decode()

    def test_literal_raw_term(self):
        kwargs = {
            'query': {
                '__0_{}'.format('field_1'): '`term_1`',
                '__result': '0'
            },
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 1)
        self.assertEqual(len(scoped_query.scope_entry_list), 2)

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, True
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, 'term_1'
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].result_scope_identifier,
            '0'
        )

    def test_scoped_term_empty(self):
        kwargs = {
            'query': {
                '__0_{}'.format('field_1'): '""',
                '__result': '0'
            },
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)
        scoped_query = search_interpreter.do_query_decode()

        self.assertEqual(len(scoped_query.get_scope_identifier_list()), 1)
        self.assertEqual(len(scoped_query.scope_entry_list), 2)

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[0], ScopedQueryEntryDataFilter
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].field_name,
            'field_1'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_quoted_value, True
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].is_raw_value, False
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].scope_identifier, '0'
        )
        self.assertEqual(
            scoped_query.scope_entry_list[0].value, ''
        )

        self.assertTrue(
            isinstance(
                scoped_query.scope_entry_list[1],
                ScopedQueryEntryControlResult
            )
        )
        self.assertEqual(
            scoped_query.scope_entry_list[1].result_scope_identifier,
            '0'
        )


class SearchInterpreterScopedDetectionTestCase(
    SearchInterpreterTestMixin, SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_subclass_scoped_selection_from_single_field(self):
        kwargs = {
            'query': {
                '__0_{}'.format(self._test_search_field.field_name): 'term1'
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        self.assertTrue(
            isinstance(search_interpreter, SearchInterpreterScoped)
        )

    def test_subclass_scoped_selection_with_single_marker_key(self):
        kwargs = {
            'query': {
                '_search_model_pk': 'test search model name',
                '__0_{}'.format(self._test_search_field.field_name): 'term1'
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        self.assertTrue(
            isinstance(search_interpreter, SearchInterpreterScoped)
        )


class SearchInterpreterScopedErrorCheckTestCase(
    SearchInterpreterTestMixin, SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_incomplete_scope_operator(self):
        kwargs = {
            'query': {
                '__0_{}'.format(self._test_search_field.field_name): 'term1',
                '__1_{}'.format(self._test_search_field.field_name): 'term1',
                '__2': 'AND',
            },
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            search_interpreter.do_query_decode()

    def test_invalid_scope_field_name(self):
        kwargs = {
            'query': {
                '__0_{}'.format(self._test_search_field.field_name): 'term1',
                '__1_INVALID': '',
                '__2': 'OR_0_1',
            },
            'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            search_interpreter.do_query_decode()

    def test_invalid_scope_operator(self):
        kwargs = {
            'query': {
                '__0_{}'.format(self._test_search_field.field_name): 'term1',
                '__1_{}'.format(self._test_search_field.field_name): 'term1',
                '__2': 'INVALID_0_1',
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            search_interpreter.do_query_decode()

    def test_invalid_scope_operator_similar(self):
        kwargs = {
            'query': {
                '__0_{}'.format(self._test_search_field.field_name): 'term1',
                '__1_{}'.format(self._test_search_field.field_name): 'term1',
                '__2': 'XOR_0_1',
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            search_interpreter.do_query_decode()

    def test_repeated_scope_id(self):
        kwargs = {
            'query': {
                '__0_{}'.format(self._test_search_field.field_name): 'term1',
                '__1_{}'.format(self._test_search_field.field_name): 'term1',
                '__1': 'OR_0_1',
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            search_interpreter.do_query_decode()

    def test_invalid_empty_value(self):
        kwargs = {
            'query': {
                '__0_{}'.format(self._test_search_field.field_name): 'term1',
                '__1_{}'.format(self._test_search_field.field_name): '',
                '__1': 'OR_0_1',
            }, 'search_model': self._test_search_model
        }

        search_interpreter = SearchInterpreter.init(**kwargs)

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            search_interpreter.do_query_decode()



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_GRAPHVIZ_DOT_PATH, DEFAULT_WORKFLOWS_IMAGE_CACHE_MAXIMUM_SIZE,
    DEFAULT_WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND,
    DEFAULT_WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS,
    DEFAULT_WORKFLOWS_WORKFLOW_STATE_ESCALATION_CHECK_INTERVAL
)
from .setting_callbacks import callback_update_workflow_image_cache_size

namespace = SettingNamespace(
    label=_('Workflows'), name='document_states'
)

setting_graphviz_dot_path = namespace.add_setting(
    default=DEFAULT_GRAPHVIZ_DOT_PATH,
    global_name='WORKFLOWS_GRAPHVIZ_DOT_PATH', help_text=_(
        'File path to the graphviz dot program used to generate workflow '
        'previews.'
    ), is_path=True
)
setting_workflow_image_cache_maximum_size = namespace.add_setting(
    default=DEFAULT_WORKFLOWS_IMAGE_CACHE_MAXIMUM_SIZE,
    global_name='WORKFLOWS_IMAGE_CACHE_MAXIMUM_SIZE',
    help_text=_(
        'The threshold at which the WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND '
        'will start deleting the oldest workflow image cache files. '
        'Specify the size in bytes.'
    ), post_edit_function=callback_update_workflow_image_cache_size
)

setting_workflow_image_cache_storage_backend = namespace.add_setting(
    default=DEFAULT_WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND,
    global_name='WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND', help_text=_(
        'Path to the Storage subclass to use when storing the cached '
        'workflow image files.'
    )
)
setting_workflow_image_cache_storage_backend_arguments = namespace.add_setting(
    default=DEFAULT_WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS,
    global_name='WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS',
    help_text=_(
        'Arguments to pass to the WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND.'
    )
)
setting_workflow_state_escalation_check_interval = namespace.add_setting(
    default=DEFAULT_WORKFLOWS_WORKFLOW_STATE_ESCALATION_CHECK_INTERVAL,
    global_name='WORKFLOWS_WORKFLOW_STATE_ESCALATION_CHECK_INTERVAL',
    help_text=_(
        'Interval in seconds on which the task to check for expired '
        'workflow states will be launched.'
    )
)



from contextlib import contextmanager
import sys


class NullFile:
    def flush(self):
        """Has no effect."""

    def write(self, string):
        """Writes here go nowhere."""


def as_id_list(items):
    return ','.join(
        [
            str(item.pk) for item in items
        ]
    )


@contextmanager
def mute_stdout():
    stdout_old = sys.stdout
    sys.stdout = NullFile()
    yield
    sys.stdout = stdout_old



from django.utils.html import format_html


def setting_widget(instance):
    return format_html(
        '<strong>{}</strong><p class="small">{}</p>', instance,
        instance.help_text or ''
    )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.sources.source_backend_actions.interface_arguments import SourceBackendActionInterfaceArgument

argument_encoded_filename = SourceBackendActionInterfaceArgument(
    help_text=_('URL safe filename of the stored file to process.')
)
argument_file_cleanup = SourceBackendActionInterfaceArgument(
    default=None, help_text=_(
        'Execute source backend specific, built-in post '
        'processing clean up code.'
    ), required=False
)
argument_file_identifier = SourceBackendActionInterfaceArgument(
    help_text=_(
        'Unique identifier to select which source backend stored file to '
        'process.'
    )
)
argument_maximum_layer_order = SourceBackendActionInterfaceArgument(
    default=None, required=False
)
argument_transformation_instance_list = SourceBackendActionInterfaceArgument(
    default=None, required=False
)
argument_user = SourceBackendActionInterfaceArgument(
    default=None, required=False
)



from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_models import Document
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from ..events import (
    event_workflow_instance_created, event_workflow_instance_transitioned
)
from ..managers import ValidWorkflowInstanceManager

from .workflow_instance_model_mixins import (
    WorkflowInstanceBusinessLogicMixin,
    WorkflowInstanceLogEntryBusinessLogicMixin
)
from .workflow_models import Workflow

__all__ = ('WorkflowInstance', 'WorkflowInstanceLogEntry')


class WorkflowInstance(
    ExtraDataModelMixin, WorkflowInstanceBusinessLogicMixin, models.Model
):
    workflow = models.ForeignKey(
        on_delete=models.CASCADE, related_name='instances', to=Workflow,
        verbose_name=_('Workflow')
    )
    datetime = models.DateTimeField(
        auto_now_add=True, db_index=True, help_text=_(
            'Workflow instance creation date time.'
        ), verbose_name=_('Datetime')
    )
    document = models.ForeignKey(
        on_delete=models.CASCADE, related_name='workflows', to=Document,
        verbose_name=_('Document')
    )
    context = models.TextField(
        blank=True, verbose_name=_('Context')
    )

    objects = models.Manager()
    valid = ValidWorkflowInstanceManager()

    class Meta:
        ordering = ('workflow',)
        unique_together = ('document', 'workflow')
        verbose_name = _('Workflow instance')
        verbose_name_plural = _('Workflow instances')

    def __str__(self):
        return str(self.workflow)

    def get_absolute_url(self):
        return reverse(
            viewname='document_states:workflow_instance_detail', kwargs={
                'workflow_instance_id': self.pk
            }
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'document',
            'event': event_workflow_instance_created,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)


class WorkflowInstanceLogEntry(
    WorkflowInstanceLogEntryBusinessLogicMixin, models.Model
):
    """
    Fields:
    * user - The user who last transitioned the document from a state to the
    Actual State.
    * datetime - Date Time - The date and time when the last user transitioned
    the document state to the Actual state.
    """
    workflow_instance = models.ForeignKey(
        on_delete=models.CASCADE, related_name='log_entries',
        to=WorkflowInstance, verbose_name=_('Workflow instance')
    )
    datetime = models.DateTimeField(
        auto_now_add=True, db_index=True, verbose_name=_('Datetime')
    )
    transition = models.ForeignKey(
        on_delete=models.CASCADE, to='WorkflowTransition',
        verbose_name=_('Transition')
    )
    user = models.ForeignKey(
        blank=True, null=True, on_delete=models.CASCADE,
        to=settings.AUTH_USER_MODEL, verbose_name=_('User')
    )
    comment = models.TextField(
        blank=True, verbose_name=_('Comment')
    )
    extra_data = models.TextField(
        blank=True, verbose_name=_('Extra data')
    )

    class Meta:
        ordering = ('datetime',)
        verbose_name = _('Workflow instance log entry')
        verbose_name_plural = _('Workflow instance log entries')

    def __str__(self):
        return str(self.transition)

    def clean(self):
        if self.transition not in self.workflow_instance.get_transition_choices(user=self.user):
            raise ValidationError(
                message=_('Not a valid transition choice.')
            )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'workflow_instance.document',
            'event': event_workflow_instance_transitioned,
            'target': 'workflow_instance'
        }
    )
    def save(self, *args, **kwargs):
        result = super().save(*args, **kwargs)
        context = self.workflow_instance.get_context()
        context.update(
            {
                'entry_log': self
            }
        )

        for action in self.transition.origin_state.exit_actions.filter(enabled=True):
            context.update(
                {
                    'action': action,
                }
            )
            action.execute(
                context=context, workflow_instance=self.workflow_instance
            )

        for action in self.transition.destination_state.entry_actions.filter(enabled=True):
            context.update(
                {
                    'action': action,
                }
            )
            action.execute(
                context=context, workflow_instance=self.workflow_instance
            )

        return result



from django.db import models, migrations
from django.conf import settings
from django.core.files.storage import FileSystemStorage


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Document',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'uuid', models.CharField(
                        default='46845ca5-d94d-4669-9d7b-7886e2bac036',
                        max_length=48, editable=False
                    )
                ),
                (
                    'label', models.CharField(
                        default='Uninitialized document',
                        help_text='The name of the document', max_length=255,
                        verbose_name='Label', db_index=True
                    )
                ),
                (
                    'description', models.TextField(
                        null=True, verbose_name='Description', blank=True
                    )
                ),
                (
                    'date_added', models.DateTimeField(
                        auto_now_add=True, verbose_name='Added'
                    )
                ),
                (
                    'language', models.CharField(
                        default='eng', max_length=8, verbose_name='Language',
                        choices=[
                            ('aar', 'Afar'), ('abk', 'Abkhazian'),
                            ('ace', 'Achinese'), ('ach', 'Acoli'),
                            ('ada', 'Adangme'), ('ady', 'Adyghe; Adygei'),
                            ('afa', 'Afro-Asiatic languages'),
                            ('afh', 'Afrihili'), ('afr', 'Afrikaans'),
                            ('ain', 'Ainu'), ('aka', 'Akan'),
                            ('akk', 'Akkadian'), ('alb', 'Albanian'),
                            ('ale', 'Aleut'), ('alg', 'Algonquian languages'),
                            ('alt', 'Southern Altai'), ('amh', 'Amharic'),
                            ('ang', 'English, Old (ca. 450-1100)'),
                            ('anp', 'Angika'), ('apa', 'Apache languages'),
                            ('ara', 'Arabic'),
                            (
                                'arc',
                                'Official Aramaic (700-300 BCE); Imperial '
                                'Aramaic (700-300 BCE)'
                            ),
                            ('arg', 'Aragonese'), ('arm', 'Armenian'),
                            ('arn', 'Mapudungun; Mapuche'), ('arp', 'Arapaho'),
                            ('art', 'Artificial languages'), ('arw', 'Arawak'),
                            ('asm', 'Assamese'),
                            ('ast', 'Asturian; Bable; Leonese; Asturleonese'),
                            ('ath', 'Athapascan languages'),
                            ('aus', 'Australian languages'), ('ava', 'Avaric'),
                            ('ave', 'Avestan'), ('awa', 'Awadhi'),
                            ('aym', 'Aymara'), ('aze', 'Azerbaijani'),
                            ('bad', 'Banda languages'),
                            ('bai', 'Bamileke languages'), ('bak', 'Bashkir'),
                            ('bal', 'Baluchi'), ('bam', 'Bambara'),
                            ('ban', 'Balinese'), ('baq', 'Basque'),
                            ('bas', 'Basa'), ('bat', 'Baltic languages'),
                            ('bej', 'Beja; Bedawiyet'), ('bel', 'Belarusian'),
                            ('bem', 'Bemba'), ('ben', 'Bengali'),
                            ('ber', 'Berber languages'), ('bho', 'Bhojpuri'),
                            ('bih', 'Bihari languages'), ('bik', 'Bikol'),
                            ('bin', 'Bini; Edo'), ('bis', 'Bislama'),
                            ('bla', 'Siksika'), ('bnt', 'Bantu languages'),
                            ('bos', 'Bosnian'), ('bra', 'Braj'),
                            ('bre', 'Breton'), ('btk', 'Batak languages'),
                            ('bua', 'Buriat'), ('bug', 'Buginese'),
                            ('bul', 'Bulgarian'), ('bur', 'Burmese'),
                            ('byn', 'Blin; Bilin'), ('cad', 'Caddo'),
                            ('cai', 'Central American Indian languages'),
                            ('car', 'Galibi Carib'),
                            ('cat', 'Catalan; Valencian'),
                            ('cau', 'Caucasian languages'),
                            ('ceb', 'Cebuano'), ('cel', 'Celtic languages'),
                            ('cha', 'Chamorro'), ('chb', 'Chibcha'),
                            ('che', 'Chechen'), ('chg', 'Chagatai'),
                            ('chi', 'Chinese'), ('chk', 'Chuukese'),
                            ('chm', 'Mari'), ('chn', 'Chinook jargon'),
                            ('cho', 'Choctaw'),
                            ('chp', 'Chipewyan; Dene Suline'),
                            ('chr', 'Cherokee'),
                            (
                                'chu',
                                'Church Slavic; Old Slavonic; Church '
                                'Slavonic; Old Bulgarian; Old Church Slavonic'
                            ),
                            ('chv', 'Chuvash'), ('chy', 'Cheyenne'),
                            ('cmc', 'Chamic languages'), ('cop', 'Coptic'),
                            ('cor', 'Cornish'), ('cos', 'Corsican'),
                            ('cpe', 'Creoles and pidgins, English based'),
                            ('cpf', 'Creoles and pidgins, French-based'),
                            ('cpp', 'Creoles and pidgins, Portuguese-based'),
                            ('cre', 'Cree'),
                            ('crh', 'Crimean Tatar; Crimean Turkish'),
                            ('crp', 'Creoles and pidgins'),
                            ('csb', 'Kashubian'),
                            ('cus', 'Cushitic languages'), ('cze', 'Czech'),
                            ('dak', 'Dakota'), ('dan', 'Danish'),
                            ('dar', 'Dargwa'),
                            ('day', 'Land Dayak languages'),
                            ('del', 'Delaware'),
                            ('den', 'Slave (Athapascan)'), ('dgr', 'Dogrib'),
                            ('din', 'Dinka'),
                            ('div', 'Divehi; Dhivehi; Maldivian'),
                            ('doi', 'Dogri'), ('dra', 'Dravidian languages'),
                            ('dsb', 'Lower Sorbian'), ('dua', 'Duala'),
                            ('dum', 'Dutch, Middle (ca. 1050-1350)'),
                            ('dut', 'Dutch; Flemish'), ('dyu', 'Dyula'),
                            ('dzo', 'Dzongkha'), ('efi', 'Efik'),
                            ('egy', 'Egyptian (Ancient)'), ('eka', 'Ekajuk'),
                            ('elx', 'Elamite'), ('eng', 'English'),
                            ('enm', 'English, Middle (1100-1500)'),
                            ('epo', 'Esperanto'), ('est', 'Estonian'),
                            ('ewe', 'Ewe'), ('ewo', 'Ewondo'),
                            ('fan', 'Fang'), ('fao', 'Faroese'),
                            ('fat', 'Fanti'), ('fij', 'Fijian'),
                            ('fil', 'Filipino; Pilipino'), ('fin', 'Finnish'),
                            ('fiu', 'Finno-Ugrian languages'), ('fon', 'Fon'),
                            ('fre', 'French'),
                            ('frm', 'French, Middle (ca. 1400-1600)'),
                            ('fro', 'French, Old (842-ca. 1400)'),
                            ('frr', 'Northern Frisian'),
                            ('frs', 'Eastern Frisian'),
                            ('fry', 'Western Frisian'), ('ful', 'Fulah'),
                            ('fur', 'Friulian'), ('gaa', 'Ga'),
                            ('gay', 'Gayo'), ('gba', 'Gbaya'),
                            ('gem', 'Germanic languages'), ('geo', 'Georgian'),
                            ('ger', 'German'), ('gez', 'Geez'),
                            ('gil', 'Gilbertese'),
                            ('gla', 'Gaelic; Scottish Gaelic'),
                            ('gle', 'Irish'), ('glg', 'Galician'),
                            ('glv', 'Manx'),
                            ('gmh', 'German, Middle High (ca. 1050-1500)'),
                            ('goh', 'German, Old High (ca. 750-1050)'),
                            ('gon', 'Gondi'), ('gor', 'Gorontalo'),
                            ('got', 'Gothic'),
                            ('grb', 'Grebo'),
                            ('grc', 'Greek, Ancient (to 1453)'),
                            ('gre', 'Greek, Modern (1453-)'),
                            ('grn', 'Guarani'),
                            ('gsw', 'Swiss German; Alemannic; Alsatian'),
                            ('guj', 'Gujarati'), ('gwi', "Gwich'in"),
                            ('hai', 'Haida'),
                            ('hat', 'Haitian; Haitian Creole'),
                            ('hau', 'Hausa'), ('haw', 'Hawaiian'),
                            ('heb', 'Hebrew'), ('her', 'Herero'),
                            ('hil', 'Hiligaynon'),
                            (
                                'him',
                                'Himachali languages; Western Pahari '
                                'languages'
                            ),
                            ('hin', 'Hindi'), ('hit', 'Hittite'),
                            ('hmn', 'Hmong; Mong'), ('hmo', 'Hiri Motu'),
                            ('hrv', 'Croatian'), ('hsb', 'Upper Sorbian'),
                            ('hun', 'Hungarian'), ('hup', 'Hupa'),
                            ('iba', 'Iban'), ('ibo', 'Igbo'),
                            ('ice', 'Icelandic'), ('ido', 'Ido'),
                            ('iii', 'Sichuan Yi; Nuosu'),
                            ('ijo', 'Ijo languages'), ('iku', 'Inuktitut'),
                            ('ile', 'Interlingue; Occidental'),
                            ('ilo', 'Iloko'),
                            (
                                'ina',
                                'Interlingua (International Auxiliary '
                                'Language Association)'
                            ),
                            ('inc', 'Indic languages'),
                            ('ind', 'Indonesian'),
                            ('ine', 'Indo-European languages'),
                            ('inh', 'Ingush'), ('ipk', 'Inupiaq'),
                            ('ira', 'Iranian languages'),
                            ('iro', 'Iroquoian languages'),
                            ('ita', 'Italian'), ('jav', 'Javanese'),
                            ('jbo', 'Lojban'), ('jpn', 'Japanese'),
                            ('jpr', 'Judeo-Persian'), ('jrb', 'Judeo-Arabic'),
                            ('kaa', 'Kara-Kalpak'), ('kab', 'Kabyle'),
                            ('kac', 'Kachin; Jingpho'),
                            ('kal', 'Kalaallisut; Greenlandic'),
                            ('kam', 'Kamba'), ('kan', 'Kannada'),
                            ('kar', 'Karen languages'), ('kas', 'Kashmiri'),
                            ('kau', 'Kanuri'), ('kaw', 'Kawi'),
                            ('kaz', 'Kazakh'), ('kbd', 'Kabardian'),
                            ('kha', 'Khasi'), ('khi', 'Khoisan languages'),
                            ('khm', 'Central Khmer'),
                            ('kho', 'Khotanese;Sakan'),
                            ('kik', 'Kikuyu; Gikuyu'), ('kin', 'Kinyarwanda'),
                            ('kir', 'Kirghiz; Kyrgyz'), ('kmb', 'Kimbundu'),
                            ('kok', 'Konkani'), ('kom', 'Komi'),
                            ('kon', 'Kongo'), ('kor', 'Korean'),
                            ('kos', 'Kosraean'), ('kpe', 'Kpelle'),
                            ('krc', 'Karachay-Balkar'), ('krl', 'Karelian'),
                            ('kro', 'Kru languages'), ('kru', 'Kurukh'),
                            ('kua', 'Kuanyama; Kwanyama'), ('kum', 'Kumyk'),
                            ('kur', 'Kurdish'), ('kut', 'Kutenai'),
                            ('lad', 'Ladino'), ('lah', 'Lahnda'),
                            ('lam', 'Lamba'), ('lao', 'Lao'),
                            ('lat', 'Latin'), ('lav', 'Latvian'),
                            ('lez', 'Lezghian'),
                            ('lim', 'Limburgan; Limburger; Limburgish'),
                            ('lin', 'Lingala'), ('lit', 'Lithuanian'),
                            ('lol', 'Mongo'), ('loz', 'Lozi'),
                            ('ltz', 'Luxembourgish; Letzeburgesch'),
                            ('lua', 'Luba-Lulua'), ('lub', 'Luba-Katanga'),
                            ('lug', 'Ganda'),
                            ('lui', 'Luiseno'), ('lun', 'Lunda'),
                            ('luo', 'Luo (Kenya and Tanzania)'),
                            ('lus', 'Lushai'), ('mac', 'Macedonian'),
                            ('mad', 'Madurese'), ('mag', 'Magahi'),
                            ('mah', 'Marshallese'), ('mai', 'Maithili'),
                            ('mak', 'Makasar'), ('mal', 'Malayalam'),
                            ('man', 'Mandingo'), ('mao', 'Maori'),
                            ('map', 'Austronesian languages'),
                            ('mar', 'Marathi'), ('mas', 'Masai'),
                            ('may', 'Malay'), ('mdf', 'Moksha'),
                            ('mdr', 'Mandar'), ('men', 'Mende'),
                            ('mga', 'Irish, Middle (900-1200)'),
                            ('mic', "Mi'kmaq; Micmac"),
                            ('min', 'Minangkabau'),
                            ('mis', 'Uncoded languages'),
                            ('mkh', 'Mon-Khmer languages'),
                            ('mlg', 'Malagasy'), ('mlt', 'Maltese'),
                            ('mnc', 'Manchu'), ('mni', 'Manipuri'),
                            ('mno', 'Manobo languages'), ('moh', 'Mohawk'),
                            ('mol', 'Moldavian; Moldovan'),
                            ('mon', 'Mongolian'), ('mos', 'Mossi'),
                            ('mul', 'Multiple languages'),
                            ('mun', 'Munda languages'), ('mus', 'Creek'),
                            ('mwl', 'Mirandese'), ('mwr', 'Marwari'),
                            ('myn', 'Mayan languages'), ('myv', 'Erzya'),
                            ('nah', 'Nahuatl languages'),
                            ('nai', 'North American Indian languages'),
                            ('nap', 'Neapolitan'), ('nau', 'Nauru'),
                            ('nav', 'Navajo; Navaho'),
                            ('nbl', 'Ndebele, South; South Ndebele'),
                            ('nde', 'Ndebele, North; North Ndebele'),
                            ('ndo', 'Ndonga'),
                            (
                                'nds',
                                'Low German; Low Saxon; German, Low; Saxon, '
                                'Low'
                            ),
                            ('nep', 'Nepali'), ('new', 'Nepal Bhasa; Newari'),
                            ('nia', 'Nias'),
                            ('nic', 'Niger-Kordofanian languages'),
                            ('niu', 'Niuean'),
                            ('nno', 'Norwegian Nynorsk; Nynorsk, Norwegian'),
                            (
                                'nob',
                                'Bokm\xe5l, Norwegian; Norwegian Bokm\xe5l'
                            ),
                            ('nog', 'Nogai'), ('non', 'Norse, Old'),
                            ('nor', 'Norwegian'),
                            ('nqo', "N'Ko"),
                            ('nso', 'Pedi; Sepedi; Northern Sotho'),
                            ('nub', 'Nubian languages'),
                            (
                                'nwc',
                                'Classical Newari; Old Newari; Classical '
                                'Nepal Bhasa'
                            ),
                            ('nya', 'Chichewa; Chewa; Nyanja'),
                            ('nym', 'Nyamwezi'),
                            ('nyn', 'Nyankole'), ('nyo', 'Nyoro'),
                            ('nzi', 'Nzima'), ('oci', 'Occitan (post 1500)'),
                            ('oji', 'Ojibwa'), ('ori', 'Oriya'),
                            ('orm', 'Oromo'), ('osa', 'Osage'),
                            ('oss', 'Ossetian; Ossetic'),
                            ('ota', 'Turkish, Ottoman (1500-1928)'),
                            ('oto', 'Otomian languages'),
                            ('paa', 'Papuan languages'),
                            ('pag', 'Pangasinan'), ('pal', 'Pahlavi'),
                            ('pam', 'Pampanga; Kapampangan'),
                            ('pan', 'Panjabi; Punjabi'),
                            ('pap', 'Papiamento'), ('pau', 'Palauan'),
                            ('peo', 'Persian, Old (ca. 600-400 B.C.)'),
                            ('per', 'Persian'),
                            ('phi', 'Philippine languages'),
                            ('phn', 'Phoenician'),
                            ('pli', 'Pali'), ('pol', 'Polish'),
                            ('pon', 'Pohnpeian'), ('por', 'Portuguese'),
                            ('pra', 'Prakrit languages'),
                            (
                                'pro',
                                'Proven\xe7al, Old (to 1500); Occitan, Old '
                                '(to 1500)'
                            ),
                            ('pus', 'Pushto; Pashto'),
                            ('qaa-qtz', 'Reserved for local use'),
                            ('que', 'Quechua'), ('raj', 'Rajasthani'),
                            ('rap', 'Rapanui'),
                            ('rar', 'Rarotongan; Cook Islands Maori'),
                            ('roa', 'Romance languages'), ('roh', 'Romansh'),
                            ('rom', 'Romany'), ('rum', 'Romanian'),
                            ('run', 'Rundi'),
                            ('rup', 'Aromanian; Arumanian; Macedo-Romanian'),
                            ('rus', 'Russian'), ('sad', 'Sandawe'),
                            ('sag', 'Sango'),
                            ('sah', 'Yakut'),
                            ('sai', 'South American Indian languages'),
                            ('sal', 'Salishan languages'),
                            ('sam', 'Samaritan Aramaic'),
                            ('san', 'Sanskrit'), ('sas', 'Sasak'),
                            ('sat', 'Santali'), ('scn', 'Sicilian'),
                            ('sco', 'Scots'), ('sel', 'Selkup'),
                            ('sem', 'Semitic languages'),
                            ('sga', 'Irish, Old (to 900)'),
                            ('sgn', 'Sign Languages'), ('shn', 'Shan'),
                            ('sid', 'Sidamo'),
                            ('sin', 'Sinhala; Sinhalese'),
                            ('sio', 'Siouan languages'),
                            ('sit', 'Sino-Tibetan languages'),
                            ('sla', 'Slavic languages'),
                            ('slo', 'Slovak'), ('slv', 'Slovenian'),
                            ('sma', 'Southern Sami'),
                            ('sme', 'Northern Sami'),
                            ('smi', 'Sami languages'),
                            ('smj', 'Lule Sami'), ('smn', 'Inari Sami'),
                            ('smo', 'Samoan'), ('sms', 'Skolt Sami'),
                            ('sna', 'Shona'), ('snd', 'Sindhi'),
                            ('snk', 'Soninke'), ('sog', 'Sogdian'),
                            ('som', 'Somali'), ('son', 'Songhai languages'),
                            ('sot', 'Sotho, Southern'),
                            ('spa', 'Spanish; Castilian'),
                            ('srd', 'Sardinian'), ('srn', 'Sranan Tongo'),
                            ('srp', 'Serbian'), ('srr', 'Serer'),
                            ('ssa', 'Nilo-Saharan languages'),
                            ('ssw', 'Swati'), ('suk', 'Sukuma'),
                            ('sun', 'Sundanese'), ('sus', 'Susu'),
                            ('sux', 'Sumerian'), ('swa', 'Swahili'),
                            ('swe', 'Swedish'), ('syc', 'Classical Syriac'),
                            ('syr', 'Syriac'),
                            ('tah', 'Tahitian'), ('tai', 'Tai languages'),
                            ('tam', 'Tamil'), ('tat', 'Tatar'),
                            ('tel', 'Telugu'), ('tem', 'Timne'),
                            ('ter', 'Tereno'), ('tet', 'Tetum'),
                            ('tgk', 'Tajik'), ('tgl', 'Tagalog'),
                            ('tha', 'Thai'), ('tib', 'Tibetan'),
                            ('tig', 'Tigre'), ('tir', 'Tigrinya'),
                            ('tiv', 'Tiv'), ('tkl', 'Tokelau'),
                            ('tlh', 'Klingon; tlhIngan-Hol'),
                            ('tli', 'Tlingit'),
                            ('tmh', 'Tamashek'), ('tog', 'Tonga (Nyasa)'),
                            ('ton', 'Tonga (Tonga Islands)'),
                            ('tpi', 'Tok Pisin'), ('tsi', 'Tsimshian'),
                            ('tsn', 'Tswana'), ('tso', 'Tsonga'),
                            ('tuk', 'Turkmen'), ('tum', 'Tumbuka'),
                            ('tup', 'Tupi languages'), ('tur', 'Turkish'),
                            ('tut', 'Altaic languages'), ('tvl', 'Tuvalu'),
                            ('twi', 'Twi'), ('tyv', 'Tuvinian'),
                            ('udm', 'Udmurt'), ('uga', 'Ugaritic'),
                            ('uig', 'Uighur; Uyghur'), ('ukr', 'Ukrainian'),
                            ('umb', 'Umbundu'), ('und', 'Undetermined'),
                            ('urd', 'Urdu'), ('uzb', 'Uzbek'),
                            ('vai', 'Vai'), ('ven', 'Venda'),
                            ('vie', 'Vietnamese'), ('vol', 'Volap\xfck'),
                            ('vot', 'Votic'), ('wak', 'Wakashan languages'),
                            ('wal', 'Wolaitta; Wolaytta'), ('war', 'Waray'),
                            ('was', 'Washo'), ('wel', 'Welsh'),
                            ('wen', 'Sorbian languages'), ('wln', 'Walloon'),
                            ('wol', 'Wolof'), ('xal', 'Kalmyk; Oirat'),
                            ('xho', 'Xhosa'), ('yao', 'Yao'),
                            ('yap', 'Yapese'), ('yid', 'Yiddish'),
                            ('yor', 'Yoruba'), ('ypk', 'Yupik languages'),
                            ('zap', 'Zapotec'),
                            ('zbl', 'Blissymbols; Blissymbolics; Bliss'),
                            ('zen', 'Zenaga'),
                            ('zgh', 'Standard Moroccan Tamazight'),
                            ('zha', 'Zhuang; Chuang'),
                            ('znd', 'Zande languages'), ('zul', 'Zulu'),
                            ('zun', 'Zuni'),
                            ('zxx', 'No linguistic content; Not applicable'),
                            (
                                'zza',
                                'Zaza; Dimili; Dimli; Kirdki; Kirmanjki; '
                                'Zazaki'
                            )
                        ]
                    )
                ),
            ],
            options={
                'ordering': ['-date_added'],
                'verbose_name': 'Document',
                'verbose_name_plural': 'Documents',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DocumentPage',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'content', models.TextField(
                        null=True, verbose_name='Content', blank=True
                    )
                ),
                (
                    'page_label', models.CharField(
                        max_length=40, null=True, verbose_name='Page label',
                        blank=True
                    )
                ),
                (
                    'page_number', models.PositiveIntegerField(
                        default=1, verbose_name='Page number',
                        editable=False, db_index=True
                    )
                ),
            ],
            options={
                'ordering': ['page_number'],
                'verbose_name': 'Document page',
                'verbose_name_plural': 'Document pages',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DocumentPageTransformation',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'order', models.PositiveIntegerField(
                        default=0, null=True, verbose_name='Order',
                        db_index=True, blank=True
                    )
                ),
                (
                    'transformation', models.CharField(
                        max_length=128, verbose_name='Transformation',
                        choices=[
                            ('resize', 'Resize'), ('rotate', 'Rotate'),
                            ('zoom', 'Zoom')
                        ]
                    )
                ),
                (
                    'arguments', models.TextField(
                        blank=True, help_text="Use dictionaries to indentify "
                        "arguments, example: {'degrees':90}", null=True,
                        verbose_name='Arguments', validators=[]
                    )
                ),
                (
                    'document_page', models.ForeignKey(
                        on_delete=models.CASCADE, to='documents.DocumentPage',
                        verbose_name='Document page'
                    )
                ),
            ],
            options={
                'ordering': ('order',),
                'verbose_name': 'Document page transformation',
                'verbose_name_plural': 'Document page transformations',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DocumentType',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'name', models.CharField(
                        unique=True, max_length=32, verbose_name='Name'
                    )
                ),
                (
                    'ocr', models.BooleanField(
                        default=True, verbose_name='Automatically queue '
                        'newly created documents for OCR.'
                    )
                ),
            ],
            options={
                'ordering': ['name'],
                'verbose_name': 'Document type',
                'verbose_name_plural': 'Documents types',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DocumentTypeFilename',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'filename', models.CharField(
                        db_index=True, max_length=128, verbose_name='Filename'
                    )
                ),
                (
                    'enabled', models.BooleanField(
                        default=True, verbose_name='Enabled'
                    )
                ),
                (
                    'document_type', models.ForeignKey(
                        on_delete=models.CASCADE, related_name='filenames',
                        to='documents.DocumentType',
                        verbose_name='Document type'
                    )
                ),
            ],
            options={
                'ordering': ['filename'],
                'verbose_name': 'Document type quick rename filename',
                'verbose_name_plural': 'Document types quick rename filenames',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DocumentVersion',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'timestamp', models.DateTimeField(
                        auto_now_add=True, verbose_name='Timestamp'
                    )
                ),
                (
                    'comment', models.TextField(
                        verbose_name='Comment', blank=True
                    )
                ),
                (
                    'file', models.FileField(
                        upload_to='2a2af9b3-8079-4753-9863-f1c342ec0d06',
                        storage=FileSystemStorage(),
                        verbose_name='File'
                    )
                ),
                (
                    'mimetype', models.CharField(
                        max_length=255, null=True, editable=False, blank=True
                    )
                ),
                (
                    'encoding', models.CharField(
                        max_length=64, null=True, editable=False, blank=True
                    )
                ),
                (
                    'checksum', models.TextField(
                        verbose_name='Checksum', null=True, editable=False,
                        blank=True
                    )
                ),
                (
                    'document', models.ForeignKey(
                        on_delete=models.CASCADE, related_name='versions',
                        to='documents.Document', verbose_name='Document'
                    )
                ),
            ],
            options={
                'verbose_name': 'Document version',
                'verbose_name_plural': 'Document version',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='RecentDocument',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'datetime_accessed', models.DateTimeField(
                        auto_now=True, verbose_name='Accessed', db_index=True
                    )
                ),
                (
                    'document', models.ForeignKey(
                        editable=False, on_delete=models.CASCADE,
                        to='documents.Document', verbose_name='Document'
                    )
                ),
                (
                    'user', models.ForeignKey(
                        editable=False, on_delete=models.CASCADE,
                        to=settings.AUTH_USER_MODEL, verbose_name='User'
                    )
                ),
            ],
            options={
                'ordering': ('-datetime_accessed',),
                'verbose_name': 'Recent document',
                'verbose_name_plural': 'Recent documents',
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='documenttypefilename',
            unique_together={('document_type', 'filename')},
        ),
        migrations.AddField(
            model_name='documentpage',
            name='document_version',
            field=models.ForeignKey(
                on_delete=models.CASCADE, related_name='pages',
                to='documents.DocumentVersion',
                verbose_name='Document version'
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='document',
            name='document_type',
            field=models.ForeignKey(
                on_delete=models.CASCADE, related_name='documents',
                to='documents.DocumentType', verbose_name='Document type'
            ),
            preserve_default=True,
        ),
    ]



from unittest import skip

from django.test import override_settings

from mayan.apps.testing.tests.base import BaseTestCase

from .mixins import (
    LockBackendTestCaseMixin, LockBackendTestMixin, DefaultTimeoutTestMixin
)


class FileLockBackendTestCase(
    LockBackendTestMixin, LockBackendTestCaseMixin, DefaultTimeoutTestMixin,
    BaseTestCase
):
    _test_locking_backend_string = 'mayan.apps.lock_manager.backends.file_lock.FileLock'


class ModelLockBackendTestCase(
    LockBackendTestMixin, LockBackendTestCaseMixin, DefaultTimeoutTestMixin,
    BaseTestCase
):
    _test_locking_backend_string = 'mayan.apps.lock_manager.backends.model_lock.ModelLock'


@skip(reason='Skip until a Mock Redis server class is added.')
@override_settings(
    LOCK_MANAGER_BACKEND_ARGUMENTS={'redis_url': 'redis://127.0.0.1:6379/0'}
)
class RedisLockBackendTestCase(
    LockBackendTestMixin, LockBackendTestCaseMixin, DefaultTimeoutTestMixin,
    BaseTestCase
):
    _test_locking_backend_string = 'mayan.apps.lock_manager.backends.redis_lock.RedisLock'



import logging

from django.contrib import messages
from django.urls import reverse, reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.common.settings import setting_home_view
from mayan.apps.views.generics import (
    ConfirmView, MultipleObjectConfirmActionView
)

from ..icons import (
    icon_document_trash_send, icon_trash_can_empty,
    icon_trashed_document_delete, icon_trashed_document_list,
    icon_trashed_document_restore
)
from ..models.document_models import Document
from ..models.trashed_document_models import TrashedDocument
from ..permissions import (
    permission_trashed_document_delete, permission_trashed_document_restore,
    permission_document_trash, permission_document_view,
    permission_trash_empty
)
from ..tasks import task_trashed_document_delete, task_trash_can_empty

from .document_views import DocumentListView

logger = logging.getLogger(name=__name__)


class DocumentTrashView(MultipleObjectConfirmActionView):
    object_permission = permission_document_trash
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid.all()
    success_message_single = _(
        'Document "%(object)s" moved to trash.'
    )
    success_message_singular = _(
        '%(count)d document moved to the trash.'
    )
    success_message_plural = _(
        '%(count)d documents moved to the trash.'
    )
    title_single = _('Move the document "%(object)s" to trash?')
    title_singular = _('Move the selected document to the trash?')
    title_plural = _('Move the %(count)d selected documents to trash?')
    view_icon = icon_document_trash_send

    def get_extra_context(self):
        context = {}

        if self.object_list.count() == 1:
            context['object'] = self.object_list.first()

        return context

    def get_post_action_redirect(self):
        # Return to the previous view after moving the document to trash
        # unless the move happened from the document view, in which case
        # redirecting back to the document is not possible because it is
        # now a trashed document and not accessible.
        if 'document_id' in self.kwargs:
            return reverse(viewname=setting_home_view.value)
        else:
            return None

    def object_action(self, form, instance):
        instance._event_actor = self.request.user
        instance.delete()


class EmptyTrashCanView(ConfirmView):
    action_cancel_redirect = post_action_redirect = reverse_lazy(
        'documents:document_list_deleted'
    )
    extra_context = {
        'title': _('Empty trash?')
    }
    view_icon = icon_trash_can_empty
    view_permission = permission_trash_empty

    def view_action(self):
        task_trash_can_empty.apply_async(
            kwargs={
                'user_id': self.request.user.pk
            }
        )

        messages.success(
            message=_('The trash emptying task has been queued.'),
            request=self.request
        )


class TrashedDocumentDeleteView(MultipleObjectConfirmActionView):
    model = TrashedDocument
    object_permission = permission_trashed_document_delete
    pk_url_kwarg = 'document_id'
    success_message_plural = _(
        '%(count)d trashed documents submitted for deletion.'
    )
    success_message_single = _(
        'Trash document "%(object)s" submitted for deletion.'
    )
    success_message_singular = _(
        '%(count)d trashed document submitted for deletion.'
    )
    title_plural = _('Delete the %(count)d selected trashed documents?')
    title_single = _('Delete the trashed document "%(object)s"?')
    title_singular = _('Delete the selected trashed document?')
    view_icon = icon_trashed_document_delete

    def get_extra_context(self):
        context = {}

        if self.object_list.count() == 1:
            context['object'] = self.object_list.first()

        return context

    def object_action(self, form, instance):
        task_trashed_document_delete.apply_async(
            kwargs={
                'trashed_document_id': instance.pk,
                'user_id': self.request.user.pk
            }
        )


class TrashedDocumentListView(DocumentListView):
    object_permission = None
    view_icon = icon_trashed_document_list

    def get_document_queryset(self):
        return AccessControlList.objects.restrict_queryset(
            permission=permission_document_view,
            queryset=TrashedDocument.trash.all(),
            user=self.request.user
        )

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'hide_link': True,
                'no_results_icon': icon_trashed_document_list,
                'no_results_text': _(
                    'To avoid loss of data, documents are not deleted '
                    'instantly. First, they are placed in the trash can. '
                    'From here they can be then finally deleted or restored.'
                ),
                'no_results_title': _(
                    'There are no documents in the trash can'
                ),
                'title': _('Documents in trash')
            }
        )
        return context


class TrashedDocumentRestoreView(MultipleObjectConfirmActionView):
    model = TrashedDocument
    object_permission = permission_trashed_document_restore
    pk_url_kwarg = 'document_id'
    success_message_plural = _(
        '%(count)d trashed documents restored.'
    )
    success_message_single = _(
        'Trashed document "%(object)s" restored.'
    )
    success_message_singular = _(
        '%(count)d trashed document restored.'
    )
    title_plural = _('Restore the %(count)d selected trashed documents?')
    title_single = _('Restore the trashed document: %(object)s')
    title_singular = _('Restore the selected trashed document?')
    view_icon = icon_trashed_document_restore

    def get_extra_context(self):
        context = {}

        if self.object_list.count() == 1:
            context['object'] = self.object_list.first()

        return context

    def object_action(self, form, instance):
        instance.restore(user=self.request.user)



import hashlib
import unittest

from fuse import FuseOSError

from django.db import connection
from django.test import tag

from mayan.apps.cabinets.tests.mixins import CabinetTestMixin
from mayan.apps.documents.tests.base import GenericDocumentTestCase

from ..filesystems import MirrorFilesystem
from ..runtime import cache

from .literals import (
    TEST_CABINET_LABEL, TEST_CABINET_LABEL_INVALID,
    TEST_CABINET_LABEL_MULTILINE, TEST_CABINET_LABEL_MULTILINE_EXPECTED,
    TEST_CABINET_LABEL_MULTILINE_2, TEST_CABINET_LABEL_MULTILINE_2_EXPECTED
)


@tag('mirroring')
@unittest.skipIf(
    condition=connection.vendor == 'mysql',
    reason='Known to fail due to unsupported feature of database manager.'
)
class CabinetMirroringTestCase(
    CabinetTestMixin, GenericDocumentTestCase
):
    auto_create_test_cabinet = False
    auto_upload_test_document = False

    def tearDown(self):
        cache.clear_all()
        super().tearDown()

    def _get_test_filesystem(self):
        def func_document_container_node():
            return self._test_cabinet_list[0]

        return MirrorFilesystem(
            func_document_container_node=func_document_container_node,
            node_text_attribute='label'
        )

    def test_document_access(self):
        self._create_test_document_stub()
        self._create_test_cabinet(
            add_test_document=True, label=TEST_CABINET_LABEL
        )

        test_filesystem = self._get_test_filesystem()
        self.assertEqual(
            test_filesystem.access(
                '/{}/{}'.format(
                    TEST_CABINET_LABEL, self._test_document.label
                )
            ), None
        )

    def test_trashed_document_access(self):
        self._create_test_document_stub()
        self._create_test_cabinet(
            add_test_document=True, label=TEST_CABINET_LABEL
        )

        self._test_document.delete()

        test_filesystem = self._get_test_filesystem()
        with self.assertRaises(expected_exception=FuseOSError):
            self.assertEqual(
                test_filesystem.access(
                    '/{}/{}'.format(
                        TEST_CABINET_LABEL, self._test_document.label
                    )
                ), None
            )

    def test_document_access_failure(self):
        self._create_test_document_stub()
        self._create_test_cabinet(
            add_test_document=True, label=TEST_CABINET_LABEL
        )

        test_filesystem = self._get_test_filesystem()
        with self.assertRaises(expected_exception=FuseOSError):
            test_filesystem.access(
                '/{}/{}_non_valid'.format(
                    TEST_CABINET_LABEL, self._test_document.label
                )
            )

    def test_document_empty(self):
        self._create_test_cabinet(label=TEST_CABINET_LABEL)
        test_filesystem = self._get_test_filesystem()

        self._upload_test_document()
        self._test_cabinet.documents.add(self._test_document)

        # Delete the physical document file without deleting the document
        # database entry.
        document_file = self._test_document.file_latest
        document_file.file.storage.delete(document_file.file.name)

        # Document file disk size is immutable since
        # 8733e07974c9ecce4880fb1fa19a14f6416c5fe6
        # 2022-04-30.
        self.assertEqual(
            test_filesystem.getattr(
                path='/{}/{}'.format(
                    TEST_CABINET_LABEL, self._test_document.label
                )
            )['st_size'], document_file.size
        )

    def test_document_open(self):
        self._upload_test_document()
        self._create_test_cabinet(
            add_test_document=True, label=TEST_CABINET_LABEL
        )

        test_filesystem = self._get_test_filesystem()

        file_handle = test_filesystem.open(
            path='/{}/{}'.format(
                TEST_CABINET_LABEL, self._test_document.label
            ), flags='rb'
        )

        self.assertEqual(
            hashlib.sha256(
                string=test_filesystem.read(
                    fh=file_handle, offset=0, path=None,
                    size=self._test_document.file_latest.size
                )
            ).hexdigest(),
            self._test_document.file_latest.checksum
        )

    def test_multiline_indexes(self):
        self._create_test_cabinet()
        self._create_test_cabinet_child(
            label=TEST_CABINET_LABEL_MULTILINE
        )

        test_filesystem = self._get_test_filesystem()
        self.assertEqual(
            list(test_filesystem.readdir('/', ''))[2:],
            [TEST_CABINET_LABEL_MULTILINE_EXPECTED]
        )

    def test_multiline_indexes_first_and_last(self):
        self._create_test_cabinet()
        self._create_test_cabinet_child(
            label=TEST_CABINET_LABEL_MULTILINE_2
        )

        test_filesystem = self._get_test_filesystem()
        self.assertEqual(
            list(test_filesystem.readdir('/', ''))[2:],
            [TEST_CABINET_LABEL_MULTILINE_2_EXPECTED]
        )

    def test_stub_documents(self):
        self._create_test_cabinet()
        self._create_test_cabinet_child(label=TEST_CABINET_LABEL)

        test_filesystem = self._get_test_filesystem()
        self.assertEqual(
            list(
                test_filesystem.readdir('/', '')
            )[2:], [TEST_CABINET_LABEL]
        )

    def test_duplicated_documents_readdir(self):
        self._create_test_cabinet(label=TEST_CABINET_LABEL)
        self._create_test_document_stub()
        self._create_test_document_stub(label=self._test_document.label)
        self._test_cabinet.documents.add(self._test_documents[0])
        self._test_cabinet.documents.add(self._test_documents[1])

        test_filesystem = self._get_test_filesystem()
        self.assertTrue(
            '{}({})'.format(
                self._test_documents[0].label, self._test_documents[0].pk
            ) in list(
                test_filesystem.readdir('/level_1', '')
            )
        )
        self.assertTrue(
            '{}({})'.format(
                self._test_documents[0].label, self._test_documents[1].pk
            ) in list(
                test_filesystem.readdir('/level_1', '')
            )
        )

    def test_duplicated_documents_open(self):
        self._create_test_cabinet(label=TEST_CABINET_LABEL)
        self._upload_test_document()
        self._upload_test_document()
        self._test_cabinet.documents.add(self._test_documents[0])
        self._test_cabinet.documents.add(self._test_documents[1])

        test_filesystem = self._get_test_filesystem()

        test_document_1_path = '/level_1/{}({})'.format(
            self._test_documents[0].label, self._test_documents[0].pk
        )
        test_document_2_path = '/level_1/{}({})'.format(
            self._test_documents[1].label, self._test_documents[1].pk
        )

        file_handle = test_filesystem.open(
            path=test_document_1_path, flags=None
        )

        self.assertEqual(
            test_filesystem.read(
                path=None, size=-1, offset=0, fh=file_handle
            ), self._test_documents[0].file_latest.open().read()
        )

        test_filesystem.release(path=None, fh=file_handle)

        file_handle = test_filesystem.open(
            path=test_document_2_path, flags=None
        )

        self.assertEqual(
            test_filesystem.read(
                path=None, size=-1, offset=0, fh=file_handle
            ), self._test_documents[1].file_latest.open().read()
        )

        test_filesystem.release(path=None, fh=file_handle)

    def test_invalid_document_label_character_open(self):
        self._upload_test_document()
        self._create_test_cabinet(
            add_test_document=True, label=TEST_CABINET_LABEL
        )

        test_filesystem = self._get_test_filesystem()

        self._test_document.label = 'Document of 2019/12'
        self._test_document.save()

        test_document_path = '/level_1/Document of 2019_12'

        file_handle = test_filesystem.open(
            path=test_document_path, flags=None
        )
        self.assertEqual(
            test_filesystem.read(
                path=None, size=-1, offset=0, fh=file_handle
            ), self._test_document.file_latest.open().read()
        )

        test_filesystem.release(path=None, fh=file_handle)

    def test_invalid_document_label_character_readdir(self):
        self._create_test_cabinet(label=TEST_CABINET_LABEL)
        self._create_test_document_stub()
        self._test_cabinet.documents.add(self._test_document)

        test_filesystem = self._get_test_filesystem()

        self._test_document.label = 'Document of 2019/12'
        self._test_document.save()

        self.assertEqual(
            list(
                test_filesystem.readdir('/level_1', '')
            )[2], 'Document of 2019_12'
        )

    def test_invalid_directory_name_character_open(self):
        self._create_test_cabinet(
            label=TEST_CABINET_LABEL_INVALID
        )
        self._upload_test_document()
        self._test_cabinet.documents.add(self._test_document)

        test_filesystem = self._get_test_filesystem()

        test_document_path = '/level_1/{}'.format(self._test_document.label)

        file_handle = test_filesystem.open(
            path=test_document_path, flags=None
        )

        self.assertEqual(
            test_filesystem.read(
                path=None, size=-1, offset=0, fh=file_handle
            ), self._test_document.file_latest.open().read()
        )

        test_filesystem.release(path=None, fh=file_handle)

    def test_invalid_directory_name_character_readdir(self):
        self._create_test_cabinet()
        self._create_test_cabinet_child(
            label=TEST_CABINET_LABEL_INVALID
        )
        self._create_test_document_stub()
        self._test_cabinet_child.documents.add(self._test_document)

        test_filesystem = self._get_test_filesystem()
        self.assertEqual(
            list(
                test_filesystem.readdir('/', '')
            )[2], 'level_1'
        )

        self.assertEqual(
            list(
                test_filesystem.readdir('/level_1', '')
            )[2], self._test_document.label
        )



from unittest import mock

from django.core.files import File

from mayan.apps.storage.models import SharedUploadedFile
from mayan.apps.testing.tests.base import BaseTestCase

from ..events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_page_created
)
from ..models.document_file_models import DocumentFile
from ..models.document_models import Document
from ..tasks.document_tasks import task_document_upload

from .literals import (
    TEST_DOCUMENT_SMALL_CHECKSUM, TEST_FILE_SMALL_FILENAME,
    TEST_DOCUMENT_SMALL_MIMETYPE, TEST_DOCUMENT_SMALL_SIZE
)
from .mixins.document_mixins import DocumentTestMixin


class DocumentTaskTestCase(DocumentTestMixin, BaseTestCase):
    auto_upload_test_document = False

    @staticmethod
    def _test_post_document_create_callback(document, test_argument):
        return test_argument

    def test_task_document_upload(self):
        self._calculate_test_document_path()

        with open(file=self._test_document_path, mode='rb') as file_object:
            test_shared_uploaded_file = SharedUploadedFile.objects.create(
                file=File(file=file_object)
            )

        self._clear_events()

        task_document_upload.apply_async(
            kwargs={
                'document_type_id': self._test_document_type.pk,
                'shared_uploaded_file_id': test_shared_uploaded_file.pk,
                'user_id': self._test_case_user.pk
            }
        )

        self._test_document = Document.objects.first()
        self._test_document_file = self._test_document.file_latest
        self._test_document_version = self._test_document.version_active
        self._test_document_version_page = self._test_document_version.pages.first()

        self.assertEqual(
            self._test_document.document_type.label,
            self._test_document_type.label
        )
        self.assertEqual(
            self._test_document.label, TEST_FILE_SMALL_FILENAME
        )

        self.assertEqual(
            self._test_document_file.exists(), True
        )
        self.assertEqual(
            self._test_document_file.size, TEST_DOCUMENT_SMALL_SIZE
        )

        self.assertEqual(
            self._test_document_file.mimetype, TEST_DOCUMENT_SMALL_MIMETYPE
        )
        self.assertEqual(
            self._test_document_file.filename, TEST_FILE_SMALL_FILENAME
        )
        self.assertEqual(self._test_document_file.encoding, 'binary')
        self.assertEqual(
            self._test_document_file.checksum, TEST_DOCUMENT_SMALL_CHECKSUM
        )
        self.assertEqual(
            self._test_document_file.pages.count(), 1
        )

        self.assertEqual(
            self._test_document_version.pages.count(), 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        # Document created

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        # Document file created

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        # Document file edited (MIME type, page count update)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        # Document version created

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(
            events[3].target, self._test_document.version_active
        )
        self.assertEqual(events[3].verb, event_document_version_created.id)

        # Document version page created

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(
            events[4].target, self._test_document_version_page
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

    @mock.patch(target='mayan.apps.documents.tests.test_document_tasks.DocumentTaskTestCase._test_post_document_create_callback')
    def test_task_post_document_create_callback(self, mocked_callback):
        self._calculate_test_document_path()

        with open(file=self._test_document_path, mode='rb') as file_object:
            test_shared_uploaded_file = SharedUploadedFile.objects.create(
                file=File(file=file_object)
            )

        test_document_count = Document.objects.count()
        test_document_file_count = DocumentFile.objects.count()

        self._clear_events()

        callback_dict = {
            'post_document_create': {
                'dotted_path': 'mayan.apps.documents.tests.test_document_tasks.DocumentTaskTestCase',
                'function_name': '_test_post_document_create_callback',
                'kwargs': {'test_argument': 'test_value'}
            }
        }

        task_document_upload.apply_async(
            kwargs={
                'callback_dict': callback_dict,
                'document_type_id': self._test_document_type.pk,
                'shared_uploaded_file_id': test_shared_uploaded_file.pk,
                'user_id': self._test_case_user.pk
            }
        )

        self.assertEqual(Document.objects.count(), test_document_count + 1)
        self.assertEqual(
            DocumentFile.objects.count(), test_document_file_count + 1
        )

        self._test_document = Document.objects.first()
        self._test_document_file = self._test_document.file_latest
        self._test_document_version = self._test_document.version_active
        self._test_document_version_page = self._test_document_version.pages.first()

        self.assertEqual(mocked_callback.call_count, 1)
        mocked_callback.assert_called_with(
            document=self._test_document, test_argument='test_value'
        )

        self.assertEqual(
            self._test_document.document_type.label,
            self._test_document_type.label
        )
        self.assertEqual(
            self._test_document.label, TEST_FILE_SMALL_FILENAME
        )

        self.assertEqual(
            self._test_document_file.exists(), True
        )
        self.assertEqual(
            self._test_document_file.size, TEST_DOCUMENT_SMALL_SIZE
        )

        self.assertEqual(
            self._test_document_file.mimetype, TEST_DOCUMENT_SMALL_MIMETYPE
        )
        self.assertEqual(
            self._test_document_file.filename, TEST_FILE_SMALL_FILENAME
        )
        self.assertEqual(self._test_document_file.encoding, 'binary')
        self.assertEqual(
            self._test_document_file.checksum, TEST_DOCUMENT_SMALL_CHECKSUM
        )
        self.assertEqual(
            self._test_document_file.pages.count(), 1
        )

        self.assertEqual(
            self._test_document_version.pages.count(), 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        # Document created

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        # Document file created

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        # Document file edited (MIME type, page count update)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        # Document version created

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(
            events[3].target, self._test_document.version_active
        )
        self.assertEqual(events[3].verb, event_document_version_created.id)

        # Document version page created

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(
            events[4].target, self._test_document_version_page
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )



from mayan.apps.testing.tests.base import BaseTestCase

from ..events import event_cache_partition_purged, event_cache_purged

from .mixins import CacheTestMixin, FileCachingTaskTestMixin


class FileCachingTaskTestCase(
    FileCachingTaskTestMixin, CacheTestMixin, BaseTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

    def test_task_cache_partition_purge(self):
        self._clear_events()

        self._execute_task_cache_partition_purge()

        self.assertEqual(self._test_cache_partition.files.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_cache_partition)
        self.assertEqual(events[0].target, self._test_cache_partition)
        self.assertEqual(events[0].verb, event_cache_partition_purged.id)

    def test_task_cache_purge(self):
        self._clear_events()

        self._execute_task_cache_purge()

        self.assertEqual(self._test_cache_partition.files.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_cache)
        self.assertEqual(events[0].actor, self._test_cache_partition)
        self.assertEqual(events[0].target, self._test_cache_partition)
        self.assertEqual(events[0].verb, event_cache_partition_purged.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_cache)
        self.assertEqual(events[1].target, self._test_cache)
        self.assertEqual(events[1].verb, event_cache_purged.id)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0007_auto_20200917_0708')
    ]

    operations = [
        migrations.AlterField(
            model_name='filemetadataentry',
            name='value',
            field=models.TextField(
                blank=True, help_text='Value of the file metadata entry.',
                max_length=255, verbose_name='Value'
            )
        )
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_MAXIMUM_SIZE,
    DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND,
    DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND_ARGUMENTS
)
from .setting_callbacks import callback_update_signature_capture_cache_size

namespace = SettingNamespace(
    label=_('Signature captures'), name='signature_captures',
)


setting_signature_capture_cache_maximum_size = namespace.add_setting(
    default=DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_MAXIMUM_SIZE,
    global_name='SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_MAXIMUM_SIZE',
    help_text=_(
        'The threshold at which the SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND '
        'will start deleting the oldest signature capture cache files. '
        'Specify the size in bytes.'
    ), post_edit_function=callback_update_signature_capture_cache_size
)
setting_signature_capture_cache_storage_backend = namespace.add_setting(
    default=DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND,
    global_name='SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND',
    help_text=_(
        'Path to the Storage subclass to use when storing the cached '
        'signature capture files.'
    )
)
setting_signature_capture_cache_storage_backend_arguments = namespace.add_setting(
    default=DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND_ARGUMENTS,
    global_name='SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND_ARGUMENTS',
    help_text=_(
        'Arguments to pass to the SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND.'
    )
)



from django.utils.translation import ugettext_lazy as _

from rest_framework.reverse import reverse

from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.documents.serializers.document_serializers import (
    DocumentSerializer
)
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import (
    FilteredPrimaryKeyRelatedField, MultiKwargHyperlinkedIdentityField
)

from .models import ResolvedSmartLink, SmartLink, SmartLinkCondition


class SmartLinkConditionSerializer(serializers.HyperlinkedModelSerializer):
    smart_link_url = serializers.HyperlinkedIdentityField(
        label=_('Smart link URL'), lookup_url_kwarg='smart_link_id',
        view_name='rest_api:smartlink-detail'
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'smart_link_id',
                'lookup_url_kwarg': 'smart_link_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'smart_link_condition_id'
            }
        ), view_name='rest_api:smartlinkcondition-detail'
    )

    class Meta:
        fields = (
            'enabled', 'expression', 'foreign_document_data', 'inclusion',
            'id', 'negated', 'operator', 'smart_link_id', 'smart_link_url',
            'url'
        )
        model = SmartLinkCondition
        read_only_fields = ('id', 'smart_link_id', 'smart_link_url', 'url')


class SmartLinkDocumentTypeAddSerializer(serializers.Serializer):
    document_type = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the document type to add to the smart link.'
        ), label=_('Document type ID'), source_model=DocumentType,
        source_permission=permission_document_type_edit
    )


class SmartLinkDocumentTypeRemoveSerializer(serializers.Serializer):
    document_type = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the document type to remove from the smart link.'
        ), label=_('Document type ID'), source_model=DocumentType,
        source_permission=permission_document_type_edit
    )


class SmartLinkSerializer(serializers.HyperlinkedModelSerializer):
    conditions_url = serializers.HyperlinkedIdentityField(
        label=_('Conditions URL'), lookup_url_kwarg='smart_link_id',
        view_name='rest_api:smartlinkcondition-list'
    )
    document_types_url = serializers.HyperlinkedIdentityField(
        label=_('Document types URL'), lookup_url_kwarg='smart_link_id',
        view_name='rest_api:smartlink-document_type-list'
    )
    document_types_add_url = serializers.HyperlinkedIdentityField(
        label=_('Document types add URL'), lookup_url_kwarg='smart_link_id',
        view_name='rest_api:smartlink-document_type-add'
    )
    document_types_remove_url = serializers.HyperlinkedIdentityField(
        label=_('Document types remove URL'),
        lookup_url_kwarg='smart_link_id',
        view_name='rest_api:smartlink-document_type-remove'
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _('URL'),
                'lookup_url_kwarg': 'smart_link_id',
                'view_name': 'rest_api:smartlink-detail'
            }
        }
        fields = (
            'conditions_url', 'document_types_url', 'document_types_add_url',
            'document_types_remove_url', 'dynamic_label', 'enabled',
            'label', 'id', 'url'
        )
        model = SmartLink
        read_only_fields = (
            'conditions_url', 'document_types_url', 'document_types_add_url',
            'document_types_remove_url', 'id', 'url'
        )


class ResolvedSmartLinkDocumentSerializer(DocumentSerializer):
    resolved_smart_link_url = serializers.SerializerMethodField(
        label=_('Resolved smart link URL')
    )

    class Meta(DocumentSerializer.Meta):
        fields = DocumentSerializer.Meta.fields + (
            'resolved_smart_link_url',
        )
        read_only_fields = DocumentSerializer.Meta.fields

    def get_resolved_smart_link_url(self, instance):
        return reverse(
            viewname='rest_api:resolvedsmartlink-detail', kwargs={
                'document_id': self.context['document'].pk,
                'resolved_smart_link_id': self.context['resolved_smart_link'].pk
            }, request=self.context['request'],
            format=self.context['format']
        )


class ResolvedSmartLinkSerializer(serializers.HyperlinkedModelSerializer):
    documents_url = serializers.SerializerMethodField(
        label=_('Documents URL')
    )
    label = serializers.SerializerMethodField(
        label=_('Label')
    )
    smart_link_url = serializers.SerializerMethodField(
        label=_('Smart link URL')
    )
    url = serializers.SerializerMethodField(
        label=_('URL')
    )

    class Meta:
        fields = (
            'documents_url', 'label', 'smart_link_url', 'url'
        )
        model = ResolvedSmartLink
        read_only_fields = fields

    def get_documents_url(self, instance):
        return reverse(
            viewname='rest_api:resolvedsmartlinkdocument-list',
            kwargs={
                'document_id': self.context['document'].pk,
                'resolved_smart_link_id': instance.pk
            },
            request=self.context['request'], format=self.context['format']
        )

    def get_label(self, instance):
        return instance.get_label_for(
            document=self.context['document']
        )

    def get_smart_link_url(self, instance):
        return reverse(
            viewname='rest_api:smartlink-detail',
            kwargs={
                'smart_link_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:resolvedsmartlink-detail',
            kwargs={
                'document_id': self.context['document'].pk,
                'resolved_smart_link_id': instance.pk
            },
            request=self.context['request'], format=self.context['format']
        )



from django.contrib import messages
from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.generics import (
    ConfirmView, SingleObjectDeleteView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalContentTypeObjectViewMixin

from .icons import (
    icon_global_error_log_entry_list, icon_object_errors,
    icon_object_error_log_entry_delete, icon_object_error_log_entry_list,
    icon_object_error_log_entry_list_clear
)
from .models import GlobalErrorLogPartitionEntry
from .permissions import (
    permission_error_log_entry_delete, permission_error_log_entry_view
)


class GlobalErrorLogEntryList(SingleObjectListView):
    model = GlobalErrorLogPartitionEntry
    object_permission = permission_error_log_entry_view
    view_icon = icon_global_error_log_entry_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_object_errors,
            'no_results_text': _(
                'This view displays the error log of different objects. '
                'An empty list is a good thing.'
            ),
            'no_results_title': _(
                'There are no error log entries'
            ),
            'title': _('Global error log')
        }


class ObjectErrorLogEntryListClearView(
    ExternalContentTypeObjectViewMixin, ConfirmView
):
    external_object_permission = permission_error_log_entry_delete
    view_icon = icon_object_error_log_entry_list_clear

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(
                'Clear error log entries for: %s' % self.external_object
            )
        }

    def view_action(self):
        self.external_object.error_log.clear(user=self.request.user)
        messages.success(
            message=_('Object error log cleared successfully'),
            request=self.request
        )


class ObjectErrorLogEntryDeleteView(
    ExternalContentTypeObjectViewMixin, SingleObjectDeleteView
):
    external_object_permission = permission_error_log_entry_delete
    object_permission = permission_error_log_entry_delete
    pk_url_kwarg = 'error_log_partition_entry_id'
    view_icon = icon_object_error_log_entry_delete

    def get_extra_context(self):
        return {
            'external_object': self.external_object,
            'navigation_object_list': ('external_object', 'object'),
            'title': _('Delete error log entry: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_source_queryset(self):
        return self.external_object.error_log.all()


class ObjectErrorLogEntryListView(
    ExternalContentTypeObjectViewMixin, SingleObjectListView
):
    external_object_permission = permission_error_log_entry_view
    view_icon = icon_object_error_log_entry_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_object_errors,
            'no_results_text': _(
                'This view displays the error log of an object. '
                'An empty list is a good thing.'
            ),
            'no_results_title': _(
                'There are no error log entries'
            ),
            'object': self.external_object,
            'title': _('Error log entries for: %s' % self.external_object)
        }

    def get_source_queryset(self):
        return self.external_object.error_log.all()



from django.core import management

from ..base import CommonAppManagementCommand


class Command(management.BaseCommand):
    help = 'Performs the required steps after a version upgrade.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--no-dependencies', action='store_true', dest='no_dependencies',
            help='Don\'t download dependencies.',
        )

    def handle(self, *args, **options):
        instance = CommonAppManagementCommand()

        instance.do_perform_upgrade(
            no_dependencies=options.get('no_dependencies', False)
        )



from django.apps import apps
from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.literals import TIME_DELTA_UNIT_CHOICES
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.common.validators import YAMLValidator
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..classes import BaseDocumentFilenameGenerator
from ..events import (
    event_document_type_created, event_document_type_edited,
    event_document_type_quick_label_created,
    event_document_type_quick_label_deleted,
    event_document_type_quick_label_edited
)
from ..literals import (
    DEFAULT_DELETE_PERIOD, DEFAULT_DELETE_TIME_UNIT,
    DEFAULT_DOCUMENT_STUB_EXPIRATION_INTERVAL
)
from ..managers import DocumentTypeManager

from .document_type_model_mixins import DocumentTypeBusinessLogicMixin

__all__ = ('DocumentType', 'DocumentTypeFilename')


class DocumentType(
    DocumentTypeBusinessLogicMixin, ExtraDataModelMixin, models.Model
):
    """
    Define document types or classes to which a specific set of
    properties can be attached.
    """
    label = models.CharField(
        help_text=_('A short text identifying the document type.'),
        max_length=196, unique=True, verbose_name=_('Label')
    )
    filename_generator_backend = models.CharField(
        default=BaseDocumentFilenameGenerator.get_default(), help_text=_(
            'The class responsible for producing the actual filename used '
            'to store the uploaded documents.'
        ), max_length=224, verbose_name=_('Filename generator backend')
    )
    filename_generator_backend_arguments = models.TextField(
        blank=True, help_text=_(
            'The arguments for the filename generator backend as a '
            'YAML dictionary.'
        ), validators=[YAMLValidator()], verbose_name=_(
            'Filename generator backend arguments'
        )
    )

    # Retention policies

    trash_time_period = models.PositiveIntegerField(
        blank=True, help_text=_(
            'Amount of time after which documents of this type will be '
            'moved to the trash.'
        ), null=True, verbose_name=_('Trash time period')
    )
    trash_time_unit = models.CharField(
        blank=True, choices=TIME_DELTA_UNIT_CHOICES, null=True, max_length=12,
        verbose_name=_('Trash time unit')
    )
    delete_time_period = models.PositiveIntegerField(
        blank=True, default=DEFAULT_DELETE_PERIOD, help_text=_(
            'Amount of time after which documents of this type in the trash '
            'will be deleted.'
        ), null=True, verbose_name=_('Delete time period')
    )
    delete_time_unit = models.CharField(
        blank=True, choices=TIME_DELTA_UNIT_CHOICES,
        default=DEFAULT_DELETE_TIME_UNIT, max_length=12, null=True,
        verbose_name=_('Delete time unit')
    )
    document_stub_pruning_enabled = models.BooleanField(
        default=True, help_text=_(
            'Delete documents that do not contain any files after a '
            'configured expiration interval.'
        ), verbose_name=_('Document stub pruning')
    )
    document_stub_expiration_interval = models.PositiveBigIntegerField(
        default=DEFAULT_DOCUMENT_STUB_EXPIRATION_INTERVAL, help_text=_(
            'Time (in seconds) after which a document stub will be '
            'considered invalid and deleted, if pruning is enabled. This '
            'an optimization setting and should only be changed for '
            'specific circumstances.'
        ), verbose_name=_('Document stub expiration interval')
    )

    objects = DocumentTypeManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Document type')
        verbose_name_plural = _('Documents types')

    def __str__(self):
        return self.label

    def delete(self, *args, **kwargs):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        for document in Document.objects.filter(document_type=self):
            document.delete(to_trash=False)

        return super().delete(*args, **kwargs)

    def get_absolute_url(self):
        return reverse(
            viewname='documents:document_type_document_list', kwargs={
                'document_type_id': self.pk
            }
        )

    def natural_key(self):
        return (self.label,)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_document_type_created,
            'target': 'self'
        },
        edited={
            'event': event_document_type_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class DocumentTypeFilename(ExtraDataModelMixin, models.Model):
    """
    List of labels available to a specific document type for the
    quick rename functionality.
    """
    document_type = models.ForeignKey(
        on_delete=models.CASCADE, related_name='filenames', to=DocumentType,
        verbose_name=_('Document type')
    )
    filename = models.CharField(
        db_index=True, max_length=128, verbose_name=_('Label')
    )
    enabled = models.BooleanField(
        default=True, verbose_name=_('Enabled')
    )

    class Meta:
        ordering = ('filename',)
        unique_together = ('document_type', 'filename')
        verbose_name = _('Quick label')
        verbose_name_plural = _('Quick labels')

    def __str__(self):
        return self.filename

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_document_type_quick_label_deleted,
        target='document_type',
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_document_type_quick_label_created,
            'action_object': 'document_type',
            'target': 'self'
        },
        edited={
            'event': event_document_type_quick_label_edited,
            'action_object': 'document_type',
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0017_auto_20200330_0851'),
    ]

    operations = [
        migrations.CreateModel(
            name='IndexInstanceNodeSearchResult',
            fields=[
            ],
            options={
                'verbose_name': 'Index instance node',
                'verbose_name_plural': 'Index instance nodes',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('document_indexing.indexinstancenode',),
        ),
        migrations.AlterModelOptions(
            name='indexinstancenode',
            options={
                'verbose_name': 'Index instance node',
                'verbose_name_plural': 'Indexes instances node'
            },
        ),
    ]



import logging

from django.apps import apps
from django.contrib.auth import get_user_model
from django.db import OperationalError

from mayan.celery import app

logger = logging.getLogger(name=__name__)


@app.task(ignore_result=True)
def task_trash_can_empty(user_id=None):
    TrashedDocument = apps.get_model(
        app_label='documents', model_name='TrashedDocument'
    )

    for trashed_document in TrashedDocument.objects.all():
        task_trashed_document_delete.apply_async(
            kwargs={
                'trashed_document_id': trashed_document.pk,
                'user_id': user_id
            }
        )


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_trashed_document_delete(self, trashed_document_id, user_id=None):
    TrashedDocument = apps.get_model(
        app_label='documents', model_name='TrashedDocument'
    )
    User = get_user_model()

    try:
        if user_id:
            user = User.objects.get(pk=user_id)
        else:
            user = None
    except OperationalError as exception:
        raise self.retry(exc=exception)

    logger.debug(msg='Executing')
    try:
        trashed_document = TrashedDocument.objects.get(
            pk=trashed_document_id
        )
        trashed_document._event_actor = user
        trashed_document.delete()
    except OperationalError as exception:
        """
        Retry trashed document deletion on database OperationalError.
        On large number of documents or document with many pages, the level
        of deletions exceed the database capacity to fulfill them. This
        causes a query deadlock where one database process waits for a
        ShareLock on a transaction which itself is blocked by another
        ShareLock on the previous transaction.

        After a timeout period of this circular transaction dependency
        an OperationalError exception will be raised and the trashed
        document deletion can be retried.
        """
        raise self.retry(exc=exception)

    logger.debug(msg='Finished')



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter

from .events import event_error_log_deleted
from .managers import ErrorLogPartitionEntryManager
from .model_mixins import (
    ErrorLogPartitionEntryBusinessLogicMixin,
    StoredErrorLogBusinessLogicMixin
)


class StoredErrorLog(StoredErrorLogBusinessLogicMixin, models.Model):
    name = models.CharField(
        max_length=128, unique=True, verbose_name=_('Internal name')
    )

    class Meta:
        ordering = ('name',)
        verbose_name = _('Error log')
        verbose_name_plural = _('Error logs')

    def __str__(self):
        return str(self.app_label)


class ErrorLogPartition(models.Model):
    error_log = models.ForeignKey(
        on_delete=models.CASCADE, related_name='partitions',
        to=StoredErrorLog, verbose_name=_('Error log')
    )
    name = models.CharField(
        db_index=True, max_length=128, verbose_name=_('Internal name')
    )
    content_type = models.ForeignKey(
        on_delete=models.CASCADE, to=ContentType
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey(
        ct_field='content_type', fk_field='object_id'
    )

    class Meta:
        unique_together = (
            ('error_log', 'name'), ('error_log', 'content_type', 'object_id')
        )
        verbose_name = _('Error log partition')
        verbose_name_plural = _('Error log partitions')

    def __str__(self):
        return self.name


class ErrorLogPartitionEntry(
    ErrorLogPartitionEntryBusinessLogicMixin, models.Model
):
    error_log_partition = models.ForeignKey(
        on_delete=models.CASCADE, related_name='entries',
        to=ErrorLogPartition, verbose_name=_('Error log partition')
    )
    datetime = models.DateTimeField(
        auto_now_add=True, db_index=True, verbose_name=_('Date and time')
    )
    text = models.TextField(blank=True, null=True, verbose_name=_('Text'))

    objects = ErrorLogPartitionEntryManager()

    class Meta:
        get_latest_by = 'datetime'
        ordering = ('datetime',)
        verbose_name = _('Error log partition entry')
        verbose_name_plural = _('Error log partition entries')

    def __str__(self):
        return '{} {}'.format(self.datetime, self.text)

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_error_log_deleted,
        target='error_log_partition.content_object'
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)


class GlobalErrorLogPartitionEntry(ErrorLogPartitionEntry):
    class Meta:
        proxy = True



from django.urls import reverse

from mayan.apps.converter.links import (
    link_transformation_delete, link_transformation_edit,
    link_transformation_select
)
from mayan.apps.converter.tests.mixins import TransformationViewTestMixin
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..layers import layer_redactions
from ..links import link_redaction_list
from ..permissions import (
    permission_redaction_create, permission_redaction_delete,
    permission_redaction_edit, permission_redaction_view
)
from ..transformations import TransformationRedactionPercent

from .literals import TEST_REDACTION_ARGUMENT


class RedactionViewTestCase(
    TransformationViewTestMixin, GenericDocumentViewTestCase
):
    _test_layer = layer_redactions
    _test_transformation_argument = TEST_REDACTION_ARGUMENT
    auto_create_test_layer = False
    auto_create_test_transformation_class = False
    TestTransformationClass = TransformationRedactionPercent

    def test_redaction_delete_link_no_permissions(self):
        self._create_test_transformation()

        self.add_test_view(test_object=self._test_transformation)

        context = self.get_test_view()
        context['layer_name'] = self._test_layer.name
        context['content_object'] = self._test_transformation_object
        resolved_link = link_transformation_delete.resolve(context=context)

        self.assertEqual(resolved_link, None)

    def test_redaction_delete_link_with_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_delete
        )

        self.add_test_view(test_object=self._test_transformation)

        context = self.get_test_view()
        context['layer_name'] = self._test_layer.name
        context['content_object'] = self._test_transformation_object
        resolved_link = link_transformation_delete.resolve(context=context)

        self.assertNotEqual(resolved_link, None)

        self.assertEqual(
            resolved_link.url, reverse(
                viewname=link_transformation_delete.view, kwargs={
                    'app_label': self._test_transformation_object_content_type.app_label,
                    'model_name': self._test_transformation_object_content_type.model,
                    'object_id': self._test_transformation_object.pk,
                    'layer_name': self._test_layer.name,
                    'transformation_id': self._test_transformation.pk
                }
            )
        )

    def test_redaction_edit_link_no_permission(self):
        self._create_test_transformation()

        self.add_test_view(test_object=self._test_transformation)

        context = self.get_test_view()
        context['layer_name'] = self._test_layer.name
        context['content_object'] = self._test_transformation_object
        resolved_link = link_transformation_edit.resolve(context=context)
        self.assertEqual(resolved_link, None)

    def test_redaction_edit_link_with_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_edit
        )

        self.add_test_view(test_object=self._test_transformation)

        context = self.get_test_view()
        context['layer_name'] = self._test_layer.name
        context['content_object'] = self._test_transformation_object
        resolved_link = link_transformation_edit.resolve(context=context)
        self.assertNotEqual(resolved_link, None)

        self.assertEqual(
            resolved_link.url, reverse(
                viewname=link_transformation_edit.view, kwargs={
                    'app_label': self._test_transformation_object_content_type.app_label,
                    'model_name': self._test_transformation_object_content_type.model,
                    'object_id': self._test_transformation_object.pk,
                    'layer_name': self._test_layer.name,
                    'transformation_id': self._test_transformation.pk
                }
            )
        )

    def test_redaction_select_link_no_permission(self):
        self._create_test_transformation()

        self.add_test_view(test_object=self._test_transformation)

        context = self.get_test_view()
        context['layer_name'] = self._test_layer.name
        context['content_object'] = self._test_transformation_object
        resolved_link = link_transformation_select.resolve(context=context)

        self.assertEqual(resolved_link, None)

    def test_redaction_select_link_with_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_create
        )

        self.add_test_view(test_object=self._test_transformation)

        context = self.get_test_view()
        context['layer_name'] = self._test_layer.name
        context['content_object'] = self._test_transformation_object
        resolved_link = link_transformation_select.resolve(context=context)
        self.assertNotEqual(resolved_link, None)

        self.assertEqual(
            resolved_link.url, reverse(
                viewname=link_transformation_select.view, kwargs={
                    'app_label': self._test_transformation_object_content_type.app_label,
                    'model_name': self._test_transformation_object_content_type.model,
                    'object_id': self._test_transformation_object.pk,
                    'layer_name': self._test_layer.name
                }
            )
        )

    def test_redaction_list_link_no_permission(self):
        self._create_test_transformation()

        self.add_test_view(test_object=self._test_transformation)

        context = self.get_test_view()
        context['layer_name'] = self._test_layer.name
        context['content_object'] = self._test_transformation_object
        resolved_link = link_redaction_list.resolve(context=context)

        self.assertEqual(resolved_link, None)

    def test_redaction_list_link_with_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_view
        )

        self.add_test_view(test_object=self._test_transformation)

        context = self.get_test_view()
        context['layer_name'] = self._test_layer.name
        context['content_object'] = self._test_transformation_object
        resolved_link = link_redaction_list.resolve(context=context)
        self.assertNotEqual(resolved_link, None)

        self.assertEqual(
            resolved_link.url, reverse(
                viewname=link_redaction_list.view, kwargs={
                    'app_label': self._test_transformation_object_content_type.app_label,
                    'model_name': self._test_transformation_object_content_type.model,
                    'object_id': self._test_transformation_object.pk,
                    'layer_name': self._test_layer.name
                }
            )
        )


class RedactionLinkDisplayTestCase(
    TransformationViewTestMixin, GenericDocumentViewTestCase
):
    _test_layer = layer_redactions
    _test_transformation_argument = TEST_REDACTION_ARGUMENT
    auto_create_test_layer = False
    auto_create_test_transformation_class = False
    TestTransformationClass = TransformationRedactionPercent

    def test_redaction_delete_link_view_with_view_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_view
        )

        self._clear_events()

        response = self._request_transformation_list_view()
        self.assertContains(
            response=response, text=str(self._test_transformation_object),
            status_code=200
        )
        self.assertContains(
            response=response,
            text=self._test_transformation.get_transformation_class().label,
            status_code=200
        )
        self.assertNotContains(
            response=response,
            text=link_transformation_delete.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_delete_link_view_with_all_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_delete
        )
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_view
        )

        self._clear_events()

        response = self._request_transformation_list_view()
        self.assertContains(
            response=response, text=str(self._test_transformation_object),
            status_code=200
        )
        self.assertContains(
            response=response,
            text=self._test_transformation.get_transformation_class().label,
            status_code=200
        )
        self.assertContains(
            response=response,
            text=link_transformation_delete.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_edit_link_view_with_view_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_view
        )

        self._clear_events()

        response = self._request_transformation_list_view()
        self.assertContains(
            response=response, text=str(self._test_transformation_object),
            status_code=200
        )
        self.assertContains(
            response=response,
            text=self._test_transformation.get_transformation_class().label,
            status_code=200
        )
        self.assertNotContains(
            response=response,
            text=link_transformation_edit.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_edit_link_view_with_all_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_edit
        )
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_view
        )

        self._clear_events()

        response = self._request_transformation_list_view()
        self.assertContains(
            response=response, text=str(self._test_transformation_object),
            status_code=200
        )
        self.assertContains(
            response=response,
            text=self._test_transformation.get_transformation_class().label,
            status_code=200
        )
        self.assertContains(
            response=response,
            text=link_transformation_edit.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from rest_framework.generics import get_object_or_404

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from .models import DocumentCheckout
from .permissions import (
    permission_document_check_in, permission_document_check_in_override,
    permission_document_check_out_detail_view
)
from .serializers import (
    DocumentCheckoutSerializer, NewDocumentCheckoutSerializer
)


class APICheckedoutDocumentListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the documents that are currently checked out.
    post: Checkout a document.
    """
    def get_serializer_class(self):
        if self.request.method == 'POST':
            return NewDocumentCheckoutSerializer
        else:
            return DocumentCheckoutSerializer

    def get_source_queryset(self):
        valid_document_queryset = Document.valid.all()

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_document_check_out_detail_view,
            queryset=DocumentCheckout.objects.all(),
            user=self.request.user
        )

        return queryset.filter(document_id__in=valid_document_queryset)


class APICheckedoutDocumentView(generics.RetrieveDestroyAPIView):
    """
    get: Retrieve the details of the selected checked out document entry.
    delete: Check in a document.
    """
    lookup_url_kwarg = 'checkout_id'
    serializer_class = DocumentCheckoutSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            '_event_keep_attributes': ('_event_actor',)
        }

    def get_source_queryset(self):
        valid_document_queryset = Document.valid.all()

        if self.request.method == 'GET':
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission_document_check_out_detail_view,
                queryset=DocumentCheckout.objects.all(),
                user=self.request.user
            )

            return queryset.filter(document_id__in=valid_document_queryset)
        elif self.request.method == 'DELETE':
            check_in_queryset = AccessControlList.objects.restrict_queryset(
                permission=permission_document_check_in,
                queryset=DocumentCheckout.objects.filter(
                    user_id=self.request.user.pk
                ), user=self.request.user
            )
            check_in_override_queryset = AccessControlList.objects.restrict_queryset(
                permission=permission_document_check_in_override,
                queryset=DocumentCheckout.objects.exclude(
                    user_id=self.request.user.pk
                ), user=self.request.user
            )

            return (check_in_queryset | check_in_override_queryset).filter(
                document_id__in=valid_document_queryset.values('pk')
            )


class APIDocumentCheckoutView(
    ExternalObjectAPIViewMixin, generics.RetrieveDestroyAPIView
):
    """
    get: Retrieve the checkout details of the selected document entry.
    delete: Check in the selected document.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permissions = {
        'GET': (permission_document_check_out_detail_view,)
    }
    serializer_class = DocumentCheckoutSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            '_event_keep_attributes': ('_event_actor',)
        }

    def get_mayan_object_permissions(self):
        if self.request.method == 'DELETE':
            try:
                checkout = self.get_external_object().checkout
            except DocumentCheckout.DoesNotExist:
                return
            else:
                permission = permission_document_check_in

                if checkout.user != self.request.user:
                    permission = permission_document_check_in_override

                return permission

    def get_object(self):
        queryset = self.filter_queryset(
            queryset=self.get_source_queryset()
        )

        obj = queryset.first()

        # Trigger a 404 error if no results are found.
        if obj:
            pk = obj.pk
        else:
            pk = None

        return get_object_or_404(queryset, pk=pk)

    def get_source_queryset(self):
        return DocumentCheckout.objects.filter(
            document=self.get_external_object()
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_('User management'), name='user_management'
)

event_group_created = namespace.add_event_type(
    label=_('Group created'), name='group_created'
)
event_group_edited = namespace.add_event_type(
    label=_('Group edited'), name='group_edited'
)

event_user_created = namespace.add_event_type(
    label=_('User created'), name='user_created'
)
event_user_edited = namespace.add_event_type(
    label=_('User edited'), name='user_edited'
)

# Deprecated events. These are now handled by the authentication app.
event_user_logged_in = namespace.add_event_type(
    label=_('User logged in'), name='user_logged_in'
)
event_user_logged_out = namespace.add_event_type(
    label=_('User logged out'), name='user_logged_out'
)



from django.conf.urls import url

from mayan.apps.views.generics import SimpleView

from .views import (
    ThemeCreateView, ThemeDeleteView, ThemeEditView, ThemeListView,
    UserThemeSettingsDetailsView, UserThemeSettingsEditView
)


urlpatterns_error_pages = [
    url(
        regex=r'^errors/403/$', name='error_403', view=SimpleView.as_view(
            template_name='403.html'
        )
    ),
    url(
        regex=r'^errors/404/$', name='error_404', view=SimpleView.as_view(
            template_name='404.html'
        )
    ),
    url(
        regex=r'^errors/500/$', name='error_500', view=SimpleView.as_view(
            template_name='500.html'
        )
    )
]

urlpatterns_themes = [
    url(
        regex=r'^themes/$', name='theme_list',
        view=ThemeListView.as_view()
    ),
    url(
        regex=r'^themes/create/$', name='theme_create',
        view=ThemeCreateView.as_view()
    ),
    url(
        regex=r'^themes/(?P<theme_id>\d+)/delete/$',
        name='theme_delete', view=ThemeDeleteView.as_view()
    ),
    url(
        regex=r'^themes/(?P<theme_id>\d+)/edit/$', name='theme_edit',
        view=ThemeEditView.as_view()
    )
]

urlpatterns_user_theme_settings = [
    url(
        regex=r'^user/(?P<user_id>\d+)/theme/$',
        name='user_theme_settings_detail',
        view=UserThemeSettingsDetailsView.as_view()
    ),
    url(
        regex=r'^user/(?P<user_id>\d+)/theme/edit/$',
        name='user_theme_settings_edit',
        view=UserThemeSettingsEditView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_error_pages)
urlpatterns.extend(urlpatterns_themes)
urlpatterns.extend(urlpatterns_user_theme_settings)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_list_facet, menu_object, menu_secondary, menu_setup
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.views.column_widgets import ObjectLinkWidget

from .classes import ModelPermission
from .events import event_acl_deleted, event_acl_edited
from .links import (
    link_acl_create, link_acl_delete, link_acl_permissions,
    link_global_acl_list
)


class ACLsApp(MayanAppConfig):
    app_namespace = 'acls'
    app_url = 'acls'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.acls'
    verbose_name = _('ACLs')

    def ready(self):
        super().ready()

        AccessControlList = self.get_model(model_name='AccessControlList')
        GlobalAccessControlListProxy = self.get_model(
            model_name='GlobalAccessControlListProxy'
        )

        EventModelRegistry.register(model=AccessControlList)

        ModelCopy(model=AccessControlList).add_fields(
            field_names=(
                'content_object', 'permissions', 'role'
            )
        )

        ModelEventType.register(
            event_types=(
                event_acl_deleted, event_acl_edited
            ), model=AccessControlList
        )

        ModelPermission.register_inheritance(
            model=AccessControlList, related='content_object',
        )

        SourceColumn(
            attribute='role', is_sortable=True, source=AccessControlList
        )

        SourceColumn(
            attribute='content_object',
            is_identifier=True,
            help_text=_(
                'Object for which the access is granted. When sorting '
                'objects, only the type is used and not the actual label of '
                'the object.'
            ), include_label=True, is_sortable=True, label=_('Object'),
            sort_field='content_type', source=GlobalAccessControlListProxy,
            widget=ObjectLinkWidget
        )

        SourceColumn(
            attribute='get_permission_count', include_label=True,
            source=AccessControlList
        )

        menu_list_facet.bind_links(
            links=(
                link_acl_permissions,
            ), sources=(AccessControlList, GlobalAccessControlListProxy,)
        )
        menu_object.bind_links(
            links=(
                link_acl_delete,
            ), sources=(AccessControlList, GlobalAccessControlListProxy,)
        )
        menu_secondary.bind_links(
            links=(link_acl_create,), sources=('acls:acl_list',)
        )
        menu_setup.bind_links(
            links=(link_global_acl_list,)
        )



from django.urls import include, path, re_path

from .views import MayanOIDCAuthenticationCallbackView

passthru_urlpatterns = [
    re_path(r'^oidc/', include('mozilla_django_oidc.urls')),
]

urlpatterns = [
    # Directly override oidc/callback as ``OIDC_CALLBACK_CLASS`` is read
    # before the app loads therefore setting it in the .ready() method has
    # no effect.
    path(
        'oidc/callback/', MayanOIDCAuthenticationCallbackView.as_view(),
        name='oidc_authentication_callback'
    )
]



import logging

from django.conf import settings
from django.contrib import messages
from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.views.generic.base import RedirectView

from mayan.apps.views.generics import (
    ConfirmView, FormView, SingleObjectListView
)
from mayan.apps.views.literals import LIST_MODE_CHOICE_ITEM

from .exceptions import DynamicSearchException
from .forms import SearchForm, AdvancedSearchForm
from .icons import (
    icon_result_list, icon_search, icon_search_advanced,
    icon_search_backend_reindex, icon_search_submit
)
from .links import link_search_again
from .literals import QUERY_PARAMETER_ANY_FIELD, SEARCH_MODEL_NAME_KWARG
from .permissions import permission_search_tools
from .search_backends import SearchBackend
from .settings import setting_match_all_default_value
from .tasks import task_reindex_backend
from .view_mixins import (
    SearchModelViewMixin, SearchQueryViewMixin, SearchResultViewMixin
)

logger = logging.getLogger(name=__name__)


class SearchAgainView(SearchQueryViewMixin, RedirectView):
    query_string = True

    def get_redirect_url(self, *args, **kwargs):
        query_dict = self.get_search_query()

        search_term_any_field = query_dict.get(
            QUERY_PARAMETER_ANY_FIELD, ''
        ).strip()

        if search_term_any_field:
            self.pattern_name = 'search:search_simple'
        else:
            self.pattern_name = 'search:search_advanced'

        return super().get_redirect_url(*args, **kwargs)


class SearchBackendReindexView(ConfirmView):
    view_icon = icon_search_backend_reindex
    view_permission = permission_search_tools

    def get_extra_context(self):
        context = {
            'subtitle': _(
                'This tool erases and populates the search backend\'s '
                'internal index.'
            ),
            'title': _('Reindex search backend')
        }

        search_backend_class = SearchBackend.get_class()

        if search_backend_class.feature_reindex:
            context['message'] = _(
                'This tool is required only for some search backends. '
                'Search results will be affected while the backend is '
                'being reindexed.'
            )
        else:
            context['message'] = _(
                'The currently selected search backend does not support or '
                'requires reindexing.'
            )

        return context

    def get_post_action_redirect(self):
        return reverse(viewname='common:tools_list')

    def view_action(self):
        task_reindex_backend.apply_async()

        messages.success(
            message=_('Search backend reindexing queued.'),
            request=self.request
        )


class SearchSimpleView(SearchModelViewMixin, SearchQueryViewMixin, FormView):
    template_name = 'appearance/generic_form.html'
    view_icon = icon_search

    def get_extra_context(self):
        self.search_model = self.get_search_model()
        return {
            'form': self.get_form(),
            'form_action': reverse(
                viewname='search:search_results', kwargs={
                    SEARCH_MODEL_NAME_KWARG: self.search_model.full_name
                }
            ),
            'search_model': self.search_model,
            'submit_icon': icon_search_submit,
            'submit_label': _('Search'),
            'submit_method': 'GET',
            'title': _('Search for: %s') % self.search_model.label
        }

    def get_form(self):
        query_dict = self.get_search_query()

        search_term_any_field = query_dict.get(
            QUERY_PARAMETER_ANY_FIELD, ''
        ).strip()

        if search_term_any_field:
            return SearchForm(
                initial={
                    QUERY_PARAMETER_ANY_FIELD: search_term_any_field
                }
            )
        else:
            return SearchForm()


class SearchAdvancedView(SearchSimpleView):
    view_icon = icon_search_advanced

    def get_extra_context(self):
        context = super().get_extra_context()

        context.update(
            {
                'title': _(
                    'Advanced search for: %s'
                ) % self.search_model.label
            }
        )

        return context

    def get_form(self):
        data = self.request.GET.dict()
        data['_match_all'] = data.get(
            '_match_all', setting_match_all_default_value.value
        )

        return AdvancedSearchForm(
            data=data, search_model=self.get_search_model()
        )


class SearchResultsView(
    SearchModelViewMixin, SearchResultViewMixin, SingleObjectListView
):
    view_icon = icon_result_list

    def get_extra_context(self):
        context = {
            'hide_object': True,
            'no_results_icon': icon_search_submit,
            'no_results_main_link': link_search_again.resolve(
                context=RequestContext(
                    dict_={
                        'search_model': self.search_model
                    }, request=self.request
                )
            ),
            'no_results_text': _(
                'Try again using different terms. '
            ),
            'no_results_title': _('No search results'),
            'search_model': self.search_model,
            'subtitle': _(
                'Search query: %s'
            ) % self.search_interpreter.to_explain(),
            'title': _('Search results for: %s') % self.search_model.label
        }

        if self.search_model.list_mode == LIST_MODE_CHOICE_ITEM:
            context['list_as_items'] = True

        return context

    def get_source_queryset(self):
        try:
            return self.get_search_queryset()
        except DynamicSearchException as exception:
            if settings.DEBUG or settings.TESTING:
                raise

            messages.error(message=exception, request=self.request)
            return self.search_model.get_queryset().none()



from django.core.exceptions import ValidationError

from mayan.apps.common.serialization import yaml_dump
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..models import DocumentMetadata

from .literals import (
    TEST_DATE_INVALID, TEST_PARSER_PATH_DATE, TEST_PARSER_DATE_VALID,
    TEST_PARSER_REGULAR_EXPRESSION, TEST_PARSER_REGULAR_EXPRESSION_PATTERN,
    TEST_PARSER_REGULAR_EXPRESSION_REPLACEMENT_TEXT, TEST_VALID_DATE
)
from .mixins.metadata_type_mixins import MetadataTypeTestMixin


class MetadataTypeParsingTestCase(
    DocumentTestMixin, MetadataTypeTestMixin, BaseTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_parsing_date(self):
        self._create_test_metadata_type(
            add_test_document_type=True, extra_kwargs={
                'parser': TEST_PARSER_PATH_DATE
            }
        )

        document_metadata = DocumentMetadata(
            document=self._test_document,
            metadata_type=self._test_metadata_type,
            value=TEST_DATE_INVALID
        )

        with self.assertRaises(expected_exception=ValidationError):
            # Should return error.
            document_metadata.full_clean()
            document_metadata.save()

        # Should not return error.
        document_metadata.value = TEST_VALID_DATE
        document_metadata.full_clean()
        document_metadata.save()

        self.assertEqual(
            getattr(
                self._test_document.metadata_value_of,
                self._test_metadata_type.name
            ), TEST_PARSER_DATE_VALID
        )

    def test_parser_regular_expression(self):
        self._create_test_metadata_type(
            add_test_document_type=True, extra_kwargs={
                'parser': TEST_PARSER_REGULAR_EXPRESSION
            }
        )

        self._create_test_metadata_type(
            add_test_document_type=True, extra_kwargs={
                'parser': TEST_PARSER_REGULAR_EXPRESSION,
                'parser_arguments': yaml_dump(
                    data={
                        'pattern': TEST_PARSER_REGULAR_EXPRESSION_PATTERN,
                        'replacement': TEST_PARSER_REGULAR_EXPRESSION_REPLACEMENT_TEXT
                    }
                )
            }
        )

        document_metadata = DocumentMetadata(
            document=self._test_document,
            metadata_type=self._test_metadata_type,
            value=TEST_PARSER_REGULAR_EXPRESSION_PATTERN
        )

        document_metadata.full_clean()
        self.assertEqual(
            document_metadata.value,
            TEST_PARSER_REGULAR_EXPRESSION_REPLACEMENT_TEXT
        )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0011_sanescanner')
    ]

    operations = [
        migrations.AddField(
            model_name='sanescanner',
            name='mode',
            field=models.CharField(
                choices=[
                    ('lineart', 'Lineart'), ('monochrome', 'Monochrome'),
                    ('color', 'Color')
                ], default='color', max_length=16, verbose_name='Mode'
            )
        ),
        migrations.AddField(
            model_name='sanescanner',
            name='resolution',
            field=models.PositiveIntegerField(
                default=300,
                help_text='Sets the resolution of the scanned image in '
                'DPI (dots per inch).',
                verbose_name='Resolution'
            ),
        )
    ]



from contextlib import contextmanager
import logging
import poplib

from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.sources.exceptions import SourceException
from mayan.apps.sources.source_backends.base import SourceBackend

from .literals import DEFAULT_EMAIL_POP3_TIMEOUT
from .mixins import SourceBackendMixinEmail

logger = logging.getLogger(name=__name__)


class SourceBackendPOP3Email(SourceBackendMixinEmail, SourceBackend):
    label = _('POP3 email')

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'timeout': {
                    'class': 'django.forms.fields.IntegerField',
                    'default': DEFAULT_EMAIL_POP3_TIMEOUT,
                    'kwargs': {
                        'min_value': 0
                    },
                    'label': _('Timeout')
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _('POP3 protocol'), {
                    'fields': ('timeout',)
                }
            ),
        )

        return fieldsets

    @contextmanager
    def _get_server(self):
        logger.debug(msg='Starting POP3 email connection')
        logger.debug(
            'host: %s', self.kwargs['host']
        )
        logger.debug(
            'ssl: %s', self.kwargs['ssl']
        )

        if self.kwargs['ssl']:
            pop3_module_name = 'POP3_SSL'
        else:
            pop3_module_name = 'POP3'

        pop3_module = getattr(poplib, pop3_module_name)

        kwargs = {
            'host': self.kwargs['host'], 'port': self.kwargs['port'],
            'timeout': self.kwargs['timeout']
        }

        server = pop3_module(**kwargs)

        credential = self.get_credential()
        password = credential.get('password')
        username = credential.get('username')

        try:
            server.getwelcome()
            server.user(user=username)
            server.pass_(pswd=password)
        except Exception as exception:
            raise SourceException(
                'Error logging to POP3 server; {}'.format(
                    exception
                )
            )
        else:
            yield server
        finally:
            server.quit()

    def action_file_delete(self, message_id):
        with self._get_server() as server:
            server.dele(which=message_id)

    def action_file_get(self, message_id):
        with self._get_server() as server:
            response, message_lines, octets = server.retr(which=message_id)
            message_combined_bytes = b'\n'.join(message_lines)
            message = force_text(s=message_combined_bytes)

            yield from self.process_message(message=message)

    def get_stored_file_list(self):
        with self._get_server() as server:
            messages_info = server.list()

            logger.debug(msg='messages_info:')
            logger.debug(msg=messages_info)
            logger.debug(
                'messages count: %s', len(
                    messages_info[1]
                )
            )

            for message_info in messages_info[1]:
                message_number, message_size = message_info.split()
                message_number = int(message_number)

                logger.debug('message_number: %s', message_number)
                logger.debug('message_size: %s', message_size)

                yield message_number



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0082_populate_document_file_size')
    ]

    operations = [
        migrations.AddField(
            model_name='documenttype',
            name='document_stub_expiration_interval',
            field=models.PositiveBigIntegerField(
                default=86400, help_text='Time (in seconds) after which '
                'a document stub will be considered invalid and deleted, '
                'if pruning is enabled. This an optimization setting and '
                'should only be changed for specific circumstances.',
                verbose_name='Document stub expiration interval'
            )
        ),
        migrations.AddField(
            model_name='documenttype',
            name='document_stub_pruning_enabled',
            field=models.BooleanField(
                default=True, help_text='Delete documents that do not '
                'contain any files.', verbose_name='Document stub pruning'
            )
        )
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Task manager'), name='task_manager'
)

permission_task_view = namespace.add_permission(
    label=_('View tasks'), name='task_view'
)
